/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"

#define ACL_CHK_USR(CMD, CRED) ret = check_cmd_user(CMD,CRED,handle->nicks,line,line_count);  if (ret != 0) return ret;
#define ACL_CHK_CID(CMD, CRED) ret = check_cmd_user(CMD,CRED,handle->cids,line,line_count);   if (ret != 0) return ret;
#define ACL_CHK_IP(CMD, CRED)  ret = check_cmd_ip(CMD,CRED,handle->networks,line,line_count); if (ret != 0) return ret;

static int check_cmd_user(const char* cmd, int status, struct linked_list* list, char* line, int line_count)
{
	char* data;
	char* data_extra;
	struct access_info* info = 0;
	
	if (!strncmp(line, cmd, strlen(cmd)))
	{
		data = &line[strlen(cmd)];
		data_extra = 0;
		data[0] = '\0';
		data++;
		
		data = strip_white_space(data);
		if (!strlen(data))
		{
			hub_log(log_fatal, "ACL parse error on line %d", line_count);
			return -1;
		}
		
		info = hub_malloc(sizeof(struct access_info));
		
		if (!info)
		{
			hub_log(log_error, "ACL parse error. Out of memory!"); \
			return -1;
		}
		
		if (strncmp(cmd, "user_", 5) == 0)
		{
			data_extra = strrchr(data, ':');
			if (data_extra)
			{
				data_extra[0] = 0;
				data_extra++;
			}
		}
		
		memset(info, 0, sizeof(struct access_info));
		
		info->data1 = hub_strdup(data);
		info->data2 = data_extra ? hub_strdup(data_extra) : 0;
		info->status = status;
		list_append(list, info);
		hub_log(log_debug, "Added ACL entry for '%s' (%s) (line %d, type=%s/%d)", data, data_extra ? data_extra : "", line_count, cmd, info->status);
		return 1;
	}
	return 0;
}


static void add_ip_range(struct linked_list* list, struct ip_ban_record* info)
{
	char buf1[INET6_ADDRSTRLEN];
	char buf2[INET6_ADDRSTRLEN];
	
	if (info->lo.af == AF_INET)
	{
		inet_ntop(AF_INET, &info->lo.internal_ip_data.in.s_addr, buf1, INET6_ADDRSTRLEN);
		inet_ntop(AF_INET, &info->hi.internal_ip_data.in.s_addr, buf2, INET6_ADDRSTRLEN);
	}
	else if (info->lo.af == AF_INET6)
	{
		inet_ntop(AF_INET6, &info->lo.internal_ip_data.in6, buf1, INET6_ADDRSTRLEN);
		inet_ntop(AF_INET6, &info->hi.internal_ip_data.in6, buf2, INET6_ADDRSTRLEN);
	}
	hub_log(log_debug, "Denying access for: %s-%s", buf1, buf2);
	
	list_append(list, info);
}


static int check_ip_range(const char* lo, const char* hi, struct ip_ban_record* info)
{
	int ret1, ret2;
	
	if ((ip_is_valid_ipv4(lo) && ip_is_valid_ipv4(hi)) ||
		(ip_is_valid_ipv6(lo) && ip_is_valid_ipv6(hi)))
	{
		ret1 = ip_convert_to_binary(lo, &info->lo);
		ret2 = ip_convert_to_binary(hi, &info->hi);
		if (ret1 == -1 || ret2 == -1 || ret1 != ret2)
		{
			return -1;
		}
		return 0;
	}
	return -1;
}


static int check_ip_mask(const char* text_addr, int bits, struct ip_ban_record* info)
{
	hub_log(log_trace, "Adding rule for IP bit mask %s/%d", text_addr, bits);
	
	if (ip_is_valid_ipv4(text_addr) ||
		ip_is_valid_ipv6(text_addr))
	{
		struct ip_addr_encap addr;
		struct ip_addr_encap mask1;
		struct ip_addr_encap mask2;
		int af = ip_convert_to_binary(text_addr, &addr);  /* 192.168.1.2 */
		int maxbits = af == AF_INET6 ? 128 : 32;
		ip_mask_create_left(af, bits, &mask1);            /* 255.255.255.0 */
		ip_mask_create_right(af, maxbits - bits, &mask2); /* 0.0.0.255 */
		ip_mask_apply_AND(&addr, &mask1, &info->lo);      /* 192.168.1.0 */
		ip_mask_apply_OR(&info->lo, &mask2, &info->hi);   /* 192.168.1.255 */
		return 0;
	}
	return -1;
}


static int check_cmd_ip(const char* cmd, int status, struct linked_list* list, char* line, int line_count)
{
	char* data1;
	char* data2;
	struct ip_ban_record* info = 0;
	int cidr_bits = 0;
		
	if (!strncmp(line, cmd, strlen(cmd)))
	{
		data1 = &line[strlen(cmd)];
		data2 = 0;
		data1[0] = '\0';
		data1++;
		
		data1 = strip_white_space(data1);
		if (!strlen(data1))
		{
			hub_log(log_fatal, "ACL parse error on line %d", line_count);
			return -1;
		}
		
		info = hub_malloc(sizeof(struct ip_ban_record));
		
		if (!info)
		{
			hub_log(log_error, "ACL parse error. Out of memory!"); \
			return -1;
		}
		
		memset(info, 0, sizeof(struct ip_ban_record));
		
		/* Extract IP-range */
		data2 = strrchr(data1, '-');
		if (data2)
		{
			cidr_bits = -1;
			data2[0] = 0;
			data2++;
			
			if (check_ip_range(data1, data2, info) == -1)
			{
				hub_free(info);
				return 0;
			}
			
			add_ip_range(list, info);

			return 1;
		}
		else
		{
			/* Extract IP-bitmask */
			data2 = strrchr(data1, '/');
			if (data2)
			{
				data2[0] = 0;
				data2++;
				cidr_bits = uhub_atoi(data2);
			}
			else
			{
				cidr_bits = 128;
			}
			
			if (check_ip_mask(data1, cidr_bits, info) == -1)
			{
				hub_free(info);
				return 0;
			}
			
			add_ip_range(list, info);
			
			return 1;
		}
	}
	return 0;
}



static int acl_parse_line(char* line, int line_count, void* ptr_data)
{
	char* pos;
	struct acl_handle* handle = (struct acl_handle*) ptr_data;
	int ret;
	
	if ((pos = strchr(line, '#')) != NULL)
	{
		pos[0] = 0;
	}
	
	if (strlen(line) == 0)
		return 0;
	
	hub_log(log_trace, "acl_parse_line(): '%s'", line);
	line = strip_white_space(line);
	
	if (!strlen(line))
	{
		hub_log(log_fatal, "ACL parse error on line %d", line_count);
		return -1;
	}
	
	hub_log(log_trace, "acl_parse_line: '%s'", line);
	
	ACL_CHK_USR("user_admin", user_auth_admin);
	ACL_CHK_USR("user_op",    user_auth_operator);
	ACL_CHK_USR("user_reg",   user_auth_regular);
	ACL_CHK_USR("deny_nick",  user_not_allowed);
	ACL_CHK_USR("ban_nick",   user_banned);
	ACL_CHK_CID("ban_cid",    user_banned);
	ACL_CHK_IP ("deny_ip",    user_not_allowed);
	
	hub_log(log_error, "Unknown ACL command on line %d: '%s'", line_count, line);
	return -1;
}


int acl_initialize(struct hub_config* config, struct acl_handle* handle)
{
	int ret;
	
	assert(handle);
	memset(handle, 0, sizeof(struct acl_handle));
	
	handle->nicks    = list_create();
	handle->cids     = list_create();
	handle->networks = list_create();
	
	if (!handle->nicks || !handle->cids || !handle->networks)
	{
		hub_log(log_fatal, "acl_initialize: Out of memory");
		return -1;
	}
	
	ret = file_read_lines(config->file_acl, handle, &acl_parse_line);
	if (ret == -1)
		return -1;
	return 0;
}


static void acl_free_access_info(void* ptr)
{
	struct access_info* info = (struct access_info*) ptr;
	if (info)
	{
		hub_free(info->data1);
		hub_free(info->data2);
		hub_free(info);
	}
}


static void acl_free_ip_info(void* ptr)
{
	struct access_info* info = (struct access_info*) ptr;
	if (info)
	{
		hub_free(info);
	}
}

int acl_shutdown(struct acl_handle* handle)
{
	if (handle->nicks)
	{
		list_clear(handle->nicks, &acl_free_access_info);
		list_destroy(handle->nicks);
		handle->nicks = 0;
	}
	
	if (handle->cids)
	{
		list_clear(handle->cids, &acl_free_access_info);
		list_destroy(handle->cids);
		handle->cids = 0;
	}
	
	if (handle->networks)
	{
		list_clear(handle->networks, &acl_free_ip_info);
		list_destroy(handle->networks);
		handle->networks = 0;
	}
	return 0;
}


struct access_info* acl_get_access_info_by_name(struct acl_handle* handle, const char* name)
{
	struct access_info* info = (struct access_info*) list_get_first(handle->nicks);
	
	while (info)
	{
		if (strcasecmp(info->data1, name) == 0)
		{
			return info;
		}
		info = (struct access_info*) list_get_next(handle->nicks);
	}
	
	return NULL;
}


struct access_info* acl_get_access_info_by_cid(struct acl_handle* handle, const char* cid)
{
	struct access_info* info = (struct access_info*) list_get_first(handle->cids);
	
	while (info)
	{
		if (strcasecmp(info->data1, cid) == 0)
		{
			return info;
		}
		info = (struct access_info*) list_get_next(handle->cids);
	}
	
	return NULL;
}


int acl_is_ip_banned(struct acl_handle* handle, const char* ip_address)
{
	struct ip_addr_encap raw;
	struct ip_ban_record* info = (struct ip_ban_record*) list_get_first(handle->networks);
	ip_convert_to_binary(ip_address, &raw);
	
	while (info)
	{
		if (raw.af == info->lo.af && ip_compare(&info->lo, &raw) <= 0 && ip_compare(&raw, &info->hi) <= 0)
		{
			return 1;
		}
		info = (struct ip_ban_record*) list_get_next(handle->networks);
	}
	return 0;
}


/*
 * This will generate the same challenge to the same user, always.
 * The challenge is made up of the time of the user connected
 * seconds since the unix epoch (modulus 1 million)
 * and the SID of the user (0-1 million).
 */
const char* password_generate_challenge(struct user* user)
{
	char buf[32];
	uint64_t tiger_res[3];
	static char tiger_buf[MAX_CID_LEN+1];
	
	snprintf(buf, 32, "%d%d%d", (int) user->tm_connected, (int) user->sid, (int) user->sd);
	
	tiger((uint64_t*) buf, strlen(buf), (uint64_t*) tiger_res);
	base32_encode((unsigned char*) tiger_res, TIGERSIZE, tiger_buf);
	buf[MAX_CID_LEN] = 0;
	
	hub_log(log_trace, "Generating challenge for user %s: '%s'", user->nick, tiger_buf);
	return (const char*) tiger_buf;
}


int password_verify(struct user* user, const char* password)
{
	char buf[1024];
	struct access_info* access;
	const char* challenge;
	char raw_cid[TIGERSIZE];
	char raw_challenge[TIGERSIZE];
	char password_calc[MAX_CID_LEN+1];
	uint64_t tiger_res[3];
	
	if (!password || !user || strlen(password) != MAX_CID_LEN) return 0;

	access = acl_get_access_info_by_name(user->hub->acl, user->nick);
	if (!access || !access->data2) return 0;

	if (TIGERSIZE+TIGERSIZE+strlen(access->data2) > 1024) return 0;

	challenge = password_generate_challenge(user);

	base32_decode(challenge, (unsigned char*) raw_challenge, MAX_CID_LEN);
	base32_decode(user->cid, (unsigned char*) raw_cid, MAX_CID_LEN);
	
	memcpy(buf, raw_cid, TIGERSIZE);
	memcpy(&buf[TIGERSIZE], (char*) access->data2, strlen(access->data2));
	memcpy(&buf[TIGERSIZE+strlen(access->data2)], raw_challenge, TIGERSIZE);
	
	tiger((uint64_t*) buf, TIGERSIZE+TIGERSIZE+strlen(access->data2), (uint64_t*) tiger_res);
	base32_encode((unsigned char*) tiger_res, TIGERSIZE, password_calc);
	
	hub_log(log_trace, "Checking password %s against %s", password, password_calc);
	
	if (strcasecmp(password, password_calc) == 0) return 1;
	
	return 0;
}



