/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "uhub.h"


#ifndef INT_MAX
#define INT_MAX 0x7fffffff
#endif

#ifndef INT_MIN
#define INT_MIN (-0x7fffffff - 1)
#endif

#define CFG_APPLY_BOOLEAN(KEY, TARGET) \
	if (strcmp(KEY, key) == 0) \
	{ \
		if      (strlen(data) == 1 && (data[0] == '1')) TARGET = 1; \
		else if (strlen(data) == 1 && (data[0] == '0')) TARGET = 0; \
		else if (strncasecmp(data, "true",  4) == 0) TARGET = 1; \
		else if (strncasecmp(data, "false", 5) == 0) TARGET = 0; \
		else if (strncasecmp(data, "yes",   3) == 0) TARGET = 1; \
		else if (strncasecmp(data, "no",    2) == 0) TARGET = 0; \
		else if (strncasecmp(data, "on",    2) == 0) TARGET = 1; \
		else if (strncasecmp(data, "off",   3) == 0) TARGET = 0; \
		else\
		{ \
			hub_log(log_fatal, "Configuration error on line %d: '%s' must be either '1' or '0'", line_count, key); \
			return -1; \
		} \
		TARGET |= 0x80000000; \
		return 0; \
	}

#define CFG_APPLY_STRING(KEY, TARGET) \
	if (strcmp(KEY, key) == 0) \
	{ \
		TARGET = hub_strdup(data); \
		return 0; \
	}


#define CFG_APPLY_INTEGER(KEY, TARGET) \
	if (strcmp(KEY, key) == 0) \
	{ \
		char* endptr; \
		int val; \
		errno = 0; \
		val = strtol(data, &endptr, 10); \
		if (((errno == ERANGE && (val == INT_MAX || val == INT_MIN)) || (errno != 0 && val == 0)) || endptr == data /*|| endptr != &data[strlen(data)-1]*/) { \
			hub_log(log_fatal, "Configuration error on line %d: '%s' must be a number", line_count, key); \
			return -1; \
		} \
		TARGET = val; \
		return 0; \
	}


#define CFG_APPLY_DEFAULT_STRING(KEY, VALUE) \
{ \
	if (KEY == 0) \
		KEY = hub_strdup(VALUE); \
}

#define CFG_APPLY_DEFAULT_INTEGER(KEY, VALUE) \
{ \
	if (KEY == 0) \
		KEY = VALUE; \
}

#define CFG_APPLY_DEFAULT_BOOLEAN(KEY, VALUE) \
{ \
	if (KEY & 0x80000000) \
	{ \
		KEY = KEY & 0x000000ff; \
	} \
	else \
	{ \
		KEY = VALUE; \
	} \
}


static void apply_config_defaults(struct hub_config* config)
{
	CFG_APPLY_DEFAULT_STRING(config->server_bind_addr,       "::");
	CFG_APPLY_DEFAULT_STRING(config->hub_name,               "My uhub hub");
	CFG_APPLY_DEFAULT_STRING(config->hub_description,        "no description");
	CFG_APPLY_DEFAULT_STRING(config->file_acl,               "");
	CFG_APPLY_DEFAULT_STRING(config->file_motd,              "");
	CFG_APPLY_DEFAULT_INTEGER(config->server_port,           1511);
	CFG_APPLY_DEFAULT_INTEGER(config->server_max_fds,        1024);
	CFG_APPLY_DEFAULT_INTEGER(config->max_users,             500);
	CFG_APPLY_DEFAULT_INTEGER(config->max_recv_buffer,       4096);
	CFG_APPLY_DEFAULT_INTEGER(config->max_send_buffer,       16384);
	CFG_APPLY_DEFAULT_INTEGER(config->max_send_buffer_soft,  8192);
	CFG_APPLY_DEFAULT_BOOLEAN(config->show_banner,           1);
	CFG_APPLY_DEFAULT_BOOLEAN(config->registered_users_only, 0);
	CFG_APPLY_DEFAULT_BOOLEAN(config->chat_only,             0);
}


static int apply_config(struct hub_config* config, char* key, char* data, int line_count)
{
	CFG_APPLY_STRING("server_bind_addr",       config->server_bind_addr);
	CFG_APPLY_STRING("hub_name",               config->hub_name);
	CFG_APPLY_STRING("hub_description",        config->hub_description);
	CFG_APPLY_STRING("file_acl",               config->file_acl);
	CFG_APPLY_STRING("file_motd",              config->file_motd);
	CFG_APPLY_INTEGER("server_port",           config->server_port);
	CFG_APPLY_INTEGER("server_max_fds",        config->server_max_fds);
	CFG_APPLY_INTEGER("max_users",             config->max_users);
	CFG_APPLY_INTEGER("max_recv_buffer",       config->max_recv_buffer);
	CFG_APPLY_INTEGER("max_send_buffer",       config->max_send_buffer);
	CFG_APPLY_INTEGER("max_send_buffer_soft",  config->max_send_buffer_soft);
	CFG_APPLY_BOOLEAN("show_banner",           config->show_banner);
	CFG_APPLY_BOOLEAN("chat_only",             config->chat_only);
	CFG_APPLY_BOOLEAN("registered_users_only", config->registered_users_only);
	
	/* Still here -- unknown directive */
	hub_log(log_fatal, "Unknown configuration directive: '%s'", key);
	return -1;
}


void free_config(struct hub_config* config)
{
	hub_free(config->server_bind_addr);
	hub_free(config->file_motd);
	hub_free(config->file_acl);
	hub_free(config->hub_name);
	hub_free(config->hub_description);
	memset(config, 0, sizeof(struct hub_config));
}


static int config_parse_line(char* line, int line_count, void* ptr_data)
{
	char* pos;
	char* key;
	char* data;
	struct hub_config* config = (struct hub_config*) ptr_data;
	
	if ((pos = strchr(line, '#')) != NULL)
	{
		pos[0] = 0;
	}
	
	if (strlen(line) == 0) return 0;
	
	hub_log(log_trace, "config_parse_line(): '%s'", line);
	
	if ((pos = strchr(line, '=')) != NULL)
	{
		pos[0] = 0;
	}
	
	key = line;
	data = &pos[1];
	
	key = strip_white_space(key);
	data = strip_white_space(data);
	
	if (!strlen(key) || !strlen(data))
	{
		hub_log(log_fatal, "Configuration parse error on line %d", line_count);
		return -1;
	}
	
	hub_log(log_trace, "config_parse_line: '%s' => '%s'", key, data);
	
	return apply_config(config, key, data, line_count);
}


int read_config(const char* file, struct hub_config* config)
{
	int ret;
	
	memset(config, 0, sizeof(struct hub_config));
	
	ret = file_read_lines(file, config, &config_parse_line);
	if (ret == -1)
		return -1;

	apply_config_defaults(config);
	return 0;
}



