/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"


void hub_handle_message(struct hub_info* hub, struct user* u, const char* line)
{
	struct adc_message* cmd = adc_msg_parse(u, line);
	if (cmd)
	{
		switch (cmd->cmd)
		{
			case ADC_CMD_HSUP: hub_handle_support(hub, u, cmd); break;
			case ADC_CMD_HPAS: hub_handle_password(hub, u, cmd); break;
			/* case ADC_CMD_HDSC: on_kick(hub, u, cmd); break; */
			case ADC_CMD_BINF: hub_handle_info(hub, u, cmd); break;
			
			/* These below aren't allowed in chat only hubs */
			case ADC_CMD_BSCH:
			case ADC_CMD_DSCH:
			case ADC_CMD_DRES:
			case ADC_CMD_ESCH:
			case ADC_CMD_FSCH:
			case ADC_CMD_DRCM:
			case ADC_CMD_DCTM:
				if (hub->config->chat_only && u->credentials < cred_operator)
				{
					break;
				}
				
			case ADC_CMD_EMSG:
			case ADC_CMD_DMSG:
			case ADC_CMD_BMSG:
			case ADC_CMD_FMSG:
			
			default:
				if (u->state == state_normal)
				{
					route_message(hub, u, cmd);
				}
				else
				{
					user_disconnect(u, quit_protocol_error);
				}
				break;
		}
		adc_msg_free(cmd);
	}
	else
	{
		if (u->state != state_normal)
		{
			user_disconnect(u, quit_protocol_error);
		}
	}
}




void hub_handle_support(struct hub_info* hub, struct user* u, struct adc_message* cmd)
{
	int index = 0;
	int ok = 1;
	char* arg = adc_msg_get_argument(cmd, index);
	
	while (arg)
	{
	
		if (strlen(arg) == 6)
		{
			fourcc_t fourcc = FOURCC(arg[2], arg[3], arg[4], arg[5]);
			if (strncmp(arg, "AD", 2) == 0)
			{
				user_support_add(u, fourcc);
			}
			else if (strncmp(arg, "RM", 2) == 0)
			{
				user_support_remove(u, fourcc);
			}
			else
			{
				ok = 0;
			}
		}
		else
		{
			ok = 0;
		}
		
		index++;
		hub_free(arg);
		arg = adc_msg_get_argument(cmd, index);
	}
	
	if (u->state == state_none)
	{
	
		if (index == 0) ok = 0; /* Need to support *SOMETHING*, at least BASE */
	
		if (ok)
		{
			hub_send_handshake(hub, u);
		}
		else
		{
			/* disconnect user. Do not send crap during initial handshake! */
			user_disconnect(u, quit_logon_error);
		}
	}
}


void hub_handle_password(struct hub_info* hub, struct user* u, struct adc_message* cmd)
{
	char* password = adc_msg_get_argument(cmd, 0);
	
	if (password_verify(u, password))
	{
		route_message(hub, u, u->info);
		on_login_success(hub, u, 1);
	}
	else
	{
		on_login_failure(hub, u, hub->command_password_wrong, "Password incorrect");
	}
	
	hub_free(password);
}


void on_kick(struct user* u, struct adc_message* cmd)
{
	hub_log(log_error, "on_kick() not implemented");
}


void update_user_info(struct user* u, struct adc_message* cmd)
{
	char prefix[2];
	char* start;
	char* end;
	char* argument;
	size_t length;
	struct adc_message* cmd_new = adc_msg_copy(u->info);
	
	if (!cmd_new) return;
	adc_msg_incref(cmd_new);
	
	hub_log(log_trace, "update_user_info(): %s. New stuff: '%s'", u->nick, cmd->cache);
		
	start = strchr(&cmd->cache[cmd->arg_offset-1], ' ');
	while (start && strlen(start) > 2)
	{
		prefix[0] = start[1];
		prefix[1] = start[2];
		start = &start[3];
		
		end = strchr(start, ' ');
		if (end == NULL)
			length = strlen(start);
		else
			length = &end[0] - &start[0];
		
		argument = hub_malloc(length + 1);
		if (argument == NULL)
		{
			adc_msg_free(cmd_new);
			cmd_new = 0;
			break;
		}
			
		memcpy(argument, start, length);
		argument[length] = 0;
		
		adc_msg_replace_named_argument(cmd_new, prefix, argument);
		hub_free(argument);

		if (!end) break;
		start = strchr(end, ' ');
	}
	
	if (cmd_new)
	{
		user_set_info(u, cmd_new);
		adc_msg_free(cmd_new);
	}
}


static int set_feature_cast_supports(struct user* u, struct adc_message* cmd)
{
	char *it, *tmp;
	
	if (adc_msg_has_named_argument(cmd, "SU"))
	{
		tmp = adc_msg_get_named_argument(cmd, "SU");
	
		if (u->feature_cast)
		{
			list_clear(u->feature_cast, &hub_free);
			list_destroy(u->feature_cast);
			u->feature_cast = 0;
		}
		u->feature_cast = list_create();
		
		it = tmp;
		while (strlen(it) > 4)
		{
			list_append(u->feature_cast, hub_strndup(it, 4));
			it = &it[5];
		}
		
		if (strlen(it) > 0)
		{
			list_append(u->feature_cast, hub_strdup(it));
		}
		
		hub_free(tmp);
	}
	return 0;
}


void hub_handle_info(struct hub_info* hub, struct user* u, struct adc_message* cmd)
{
	char* tmp;
	char* tmp2;
	int want_ipv4, want_ipv6;
	struct access_info* access = 0;
	int need_auth = 0;
	
	adc_msg_remove_named_argument(cmd, "HU");
	adc_msg_remove_named_argument(cmd, "HI");
	adc_msg_remove_named_argument(cmd, "OP");
	adc_msg_remove_named_argument(cmd, "BO");
	adc_msg_remove_named_argument(cmd, "TO");
	
	set_feature_cast_supports(u, cmd);
	
	if (u->state == state_protocol)
	{
		assert(u->info == NULL);
		
		if (adc_msg_has_named_argument(cmd, "NI"))
		{
			tmp = adc_msg_get_named_argument(cmd, "NI");
			memcpy(u->nick, tmp, MIN(strlen(tmp), MAX_NICK_LEN));
			hub_free(tmp);
		}
		
		if (adc_msg_has_named_argument(cmd, "ID"))
		{
			tmp = adc_msg_get_named_argument(cmd, "ID");
			memcpy(u->cid, tmp, MIN(strlen(tmp), MAX_CID_LEN));
			hub_free(tmp);
		}
		
		if (adc_msg_has_named_argument(cmd, "VE"))
		{
			tmp = adc_msg_get_named_argument(cmd, "VE");
			if (tmp)
			{
				tmp2 = adc_msg_unescape(tmp);
				if (tmp2)
				{
					memcpy(u->user_agent, tmp2, MIN(strlen(tmp2), MAX_UA_LEN));
					hub_free(tmp2);
				}
			}
			else
			{
				memcpy(u->user_agent, "unknown", 7);
			}
			hub_free(tmp);
		}
		else
		{
			memcpy(u->user_agent, "unknown", 7);
		}
		
		if (!hub_handle_info_check_nick(hub, u, cmd))
		{
			return;
		}
		
		if (!hub_handle_info_check_cid(hub, u, cmd))
		{
			on_login_failure(hub, u, NULL, "cid/pid invalid");
			return;
		}
		
		adc_msg_remove_named_argument(cmd, "PD");

		/* Set initial info */
		user_set_info(u, cmd);
		
		/* Strip IP addresses */
		adc_msg_remove_named_argument(cmd, "U6");
		want_ipv6 = adc_msg_remove_named_argument(cmd, "I6");
		adc_msg_remove_named_argument(cmd, "U4");
		want_ipv4 = adc_msg_remove_named_argument(cmd, "I4");
		
		/* Add correct/verified IP addresses instead (if requested/stripped) */
		tmp = (char*) net_get_peer_address(u->sd);
		if (tmp)
		{
			if (want_ipv4 && strchr(tmp, '.'))
			{
				adc_msg_add_named_argument(cmd, "I4", tmp);
			}
			else if (want_ipv6)
			{
				adc_msg_add_named_argument(cmd, "I6", tmp);
			}
		}
		
		tmp = adc_msg_get_named_argument(cmd, "NI");
		if (!tmp)
		{
			on_login_failure(hub, u, NULL, "Out of memory");
			return;
		}
		
		access = acl_get_access_info_by_name(hub->acl, tmp);
		hub_free(tmp);
		
		/*
		 * Check for any access rules associated with the nickname.
		 */
		if (access)
		{
			switch (access->status)
			{
				case user_banned:
					on_login_failure(hub, u, hub->command_ban_perm, "Banned by nickname");
					return;
					break;
					
				case user_not_allowed:
					on_login_failure(hub, u, hub->command_error_nick_invalid, "Nickname is not allowed");
					return;
					break;
					
				case user_not_found:
					u->credentials = cred_user;
					if (hub->config->registered_users_only)
					{
						on_login_failure(hub, u, hub->command_registered_only, "Registered users only");
						return;
					}
					break;
					
				case user_auth_regular:
					u->credentials = cred_user_registered;
					adc_msg_add_named_argument(cmd, "RG", "1");
					need_auth = 1;
					break;
					
				case user_auth_operator:
					u->credentials = cred_operator;
					adc_msg_add_named_argument(cmd, "OP", "1");
					need_auth = 1;
					break;
					
				case user_auth_admin:
					u->credentials = cred_admin;
					adc_msg_add_named_argument(cmd, "OP", "1");
					need_auth = 1;
					break;
			}
		}
		
		tmp = adc_msg_get_named_argument(cmd, "ID");
		if (!tmp)
		{
			on_login_failure(hub, u, NULL, "Out of memory");
			return;
		}
		
		/*
		 * Check if CID is banned.
		 */
		access = acl_get_access_info_by_cid(hub->acl, tmp);
		hub_free(tmp);
		if (access && access->status == user_banned)
		{
			on_login_failure(hub, u, hub->command_ban_perm, "Banned by CID");
			return;
		}
		
		/*
		 * If hub is full, don't let users in.
		 * We still allows operators and admins to enter.
		 */
		if (list_size(hub->user_list) >= hub->config->max_users && u->credentials < user_auth_operator)
		{
			on_login_failure(hub, u, hub->command_error_hub_full, "Hub full");
			return;
		}
		
		/*
		 * If user needs to be authenticated (using password),
		 * send the challenge.
		 *
		 * Otherwise, welcome the user.
		 */
		if (need_auth)
		{
			hub_send_password_challenge(hub, u);
		}
		else
		{
			route_message(hub, u, cmd);
			on_login_success(hub, u, 0);
		}
	}
	else
	{
		/* These must not be allowed #pdated */
		adc_msg_remove_named_argument(cmd, "PD");
		adc_msg_remove_named_argument(cmd, "ID");
		adc_msg_remove_named_argument(cmd, "U6");
		adc_msg_remove_named_argument(cmd, "I6");
		adc_msg_remove_named_argument(cmd, "U4");
		adc_msg_remove_named_argument(cmd, "I4");
		
		
#if ALLOW_NICK_CHANGES
		/*
		 * If the nick is not accepted, do not relay it.
		 * Otherwise, the nickname will be updated.
		 */
		if (adc_msg_has_named_argument(cmd, "NI")) {
			if (!hub_handle_info_check_nick(hub, u, cmd))
			{
				adc_msg_remove_named_argument(cmd, "NI");
			}
		}
#else
		adc_msg_remove_named_argument(cmd, "NI");
#endif
		
		update_user_info(u, cmd);
		route_message(hub, u, cmd);
	}
	
	
}


void hub_send_support(struct hub_info* hub, struct user* u)
{
	route_to_user(u, hub->command_support);
	u->state = state_protocol;
}


void hub_send_sid(struct hub_info* hub, struct user* u)
{
	struct adc_message* command;
	
	command = adc_msg_create("ISID");
	adc_msg_add_argument(command, (const char*) sid_to_string(u->sid));
	
	route_to_user(u, command);
	adc_msg_free(command);
	
	u->state = state_protocol;
}


void hub_send_hubinfo(struct hub_info* hub, struct user* u)
{
	route_to_user(u, hub->command_info);
	
	if (hub->config->show_banner)
		route_to_user(u, hub->command_banner);
	
	u->state = state_protocol;
}


void hub_send_handshake(struct hub_info* hub, struct user* u)
{
	hub_send_support(hub, u);
	hub_send_sid(hub, u);
	hub_send_hubinfo(hub, u);
}


void hub_send_motd(struct hub_info* hub, struct user* u)
{
	if (hub->command_motd)
	{
		route_to_user(u, hub->command_motd);
	}
}


void hub_send_password_challenge(struct hub_info* hub, struct user* u)
{
	struct adc_message* igpa;
	igpa = adc_msg_create("IGPA");
	adc_msg_add_argument(igpa, password_generate_challenge(u));
	u->state = state_verify;
	route_to_user(u, igpa);
	adc_msg_free(igpa);
}



int handle_client_network(struct net_event_listener* listener)
{
	static char buf[MAX_RECV_BUF];
	char* pos;
	char* start;
	ssize_t offset;
	ssize_t size;
	ssize_t buflen;
	struct user* user = (struct user*) listener->data;
	struct hub_info* hub = user ? user->hub : NULL;
	int fd = listener->fd;
	int read_more = 1;
	
	/*
	hub_log(log_trace, "handle_client_network(): network event fired (fd=%d, event=%d)", fd, listener->revents);
	*/
	
	if (listener->revents & evt_write)
	{
		send_queued_data(user);
	}
	
	if (listener->revents & evt_read)
	{
		while (read_more)
		{
		
			offset = 0;
			if (user->recv_buf)
			{
				memcpy(buf, user->recv_buf, user->recv_buf_offset);
				offset = user->recv_buf_offset;
			}
			else
			{
				offset = 0;
			}
				
			size = net_recv(fd, &buf[offset], MAX_RECV_BUF - offset, 0);
			if (size == -1)
			{
				if (errno == EAGAIN)
				{
					read_more = 0;
					break;
				}
				
				listener->revents |= evt_error;
				read_more = 0;
			}
			else if (size == 0)
			{
				listener->revents |= evt_close;
				read_more = 0;
			}
			else
			{
				buflen = offset + size;
				start = buf;
				while ((pos = strchr(start, '\n')))
				{
					pos[0] = '\0';
					if (strlen(start) > 0 && strlen(start) < hub->config->max_recv_buffer)
					{
						hub_handle_message(hub, user, start);
					}
					start = &pos[1];
				}
				
				if (&buf[offset + size] > &start[0])
				{
					if (!user->recv_buf)
					{
						user->recv_buf = hub_malloc(hub->config->max_recv_buffer);
					}
					
					if (!user->recv_buf)
					{
						listener->revents |= evt_error;
					}
					else
					{
						memcpy(user->recv_buf, start, &buf[offset + size] - &start[0]);
						user->recv_buf_offset = &buf[offset + size] - &start[0];
					}
				}
				else
				{
					if (user->recv_buf)
					{
						hub_free(user->recv_buf);
						user->recv_buf = 0;
						user->recv_buf_offset = 0;
					}
				}
			}
		}
	}
	
	if (listener->revents & evt_close)
	{
		user_disconnect(user, quit_disconnected);
	}
	else if (listener->revents & evt_error)
	{
		user_disconnect(user, quit_socket_error);
	}
	return 0;
}


int handle_server_accept(struct net_event_listener* listener)
{
	int fd;
	struct user* new_user;
	struct hub_info* hub = (struct hub_info*) listener->data;
	
	int accept_more = 1;
	
	while (accept_more)
	{
		
		fd = net_accept(listener->fd);
		if (fd == -1)
		{
			if (errno == EAGAIN)
			{
				accept_more = 0;
				break;
			}
			else
			{
				return -1;
			}
		}
		
		const char* addr = net_get_peer_address(fd);
		
		/* Check if IP is banned, if so. Close immediately! */
		if (acl_is_ip_banned(hub->acl, addr))
		{
			hub_log(log_info, "Denied      [%s] (IP banned)", addr);
			net_close(fd);
			continue;
		}
		else
		{
			hub_log(log_trace, "Got connection from %s", addr);
		}
		
		new_user = user_create(hub, fd);
		if (!new_user)
		{
			hub_log(log_error, "Unable to create user after socket accepted. Out of memory?");
			net_close(fd);
			return -1;
		}
		
		net_set_nonblocking(fd, 1);
		
		if (net_add(fd, evt_read | evt_close, new_user, &handle_client_network) == -1)
		{
			return -1;
		}
	}
	return 0;
}


struct hub_info* hub_start_service(struct hub_config* config)
{
	struct hub_info* hub = 0;
	int server, ret;
	int ipv6_supported;
	struct sockaddr_storage addr;
	socklen_t sockaddr_size;
	
	hub = hub_malloc(sizeof(struct hub_info)); /* FIXME: This is leaked if we exit by error further below! */
	if (!hub)
	{
		hub_log(log_fatal, "Unable to allocate memory for hub");
		return 0;
	}
	
	hub_log(log_trace, "Checking if IPv6 is supported...");
	ipv6_supported = net_is_ipv6_supported();
	
	if (ipv6_supported)
		hub_log(log_debug, "IPv6 supported and enabled.");
	else
		hub_log(log_debug, "IPv6 is not supported.");
	
	hub_log(log_info, "Starting server, listening on %s:%d", (ipv6_supported ? SERVER_ADDR_IPV6 : SERVER_ADDR_IPV4), config->server_port);

	server = net_socket_create((ipv6_supported ? PF_INET6 : PF_INET), SOCK_STREAM, IPPROTO_TCP);
	if (server == -1)
		return 0;
	
	if (ip_convert_address(config->server_bind_addr, config->server_port, (struct sockaddr*) &addr, &sockaddr_size) == -1)
	{
		net_close(server);
		return 0;
	}

	ret = net_set_reuseaddress(server, 1);
	if (ret == -1)
	{
		net_close(server);
		return 0;
	}
	
	ret = net_set_nonblocking(server, 1);
	if (ret == -1)
	{
		net_close(server);
		return 0;
	}
	
	ret = bind(server, (struct sockaddr*) &addr, sockaddr_size);
	if (ret == -1)
	{
		hub_log(log_fatal, "hub_start_service(): Unable to bind to local address. errno=%d, str=%s", errno, strerror(errno));
		net_close(server);
		return 0;
	}

	ret = listen(server, SERVER_BACKLOG);
	if (ret == -1)
	{
		hub_log(log_fatal, "hub_start_service(): Unable to listen to socket");
		net_close(server);
		return 0;
	}
	
	if (net_add(server, evt_accept, hub, &handle_server_accept) == -1)
	{
		net_close(server);
		return 0;
	}
	
	hub->fd = server;
	hub->config = config;
	hub->user_list = NULL;
	hub->free_sid = 1;
	
	if (user_manager_init(hub) == -1)
	{
		net_remove(server);
		net_close(server);
		return 0;
	}

	
	return hub;
}


void hub_shutdown_service(struct hub_info* hub)
{
	net_remove(hub->fd);
	net_close(hub->fd);
	user_manager_shutdown(hub);
	hub_free(hub);
	hub = 0;
}


void hub_set_variables(struct hub_info* hub, struct acl_handle* acl)
{
	int fd, ret;
	char buf[MAX_RECV_BUF];
	char* tmp;
	
	hub->acl = acl;
	
	hub->command_info = adc_msg_create("IINF HU1 HI1 VE" PRODUCT "/" VERSION "");
	
	tmp = adc_msg_escape(hub->config->hub_name);
	adc_msg_add_named_argument(hub->command_info, "NI", tmp);
	hub_free(tmp);
	
	tmp = adc_msg_escape(hub->config->hub_description);
	adc_msg_add_named_argument(hub->command_info, "DE", tmp);
	hub_free(tmp);
	
	
	/* (Re-)read the message of the day */
	hub->command_motd = 0;
	fd = open(hub->config->file_motd, 0);
	if (fd != -1)
	{
		ret = read(fd, buf, MAX_RECV_BUF);
		if (ret > 0)
		{
			buf[ret] = 0;
			hub->command_motd = adc_msg_create("IMSG");
			tmp = adc_msg_escape(buf);
 			adc_msg_add_argument(hub->command_motd, tmp);
			hub_free(tmp);
		}
		else
		{
			
		}
		close(fd);
	}
	
	hub->command_support            = adc_msg_create("ISUP " ADC_PROTO_SUPPORT "");
	hub->command_error_nick_missing = adc_msg_create("ISTA 243 No\\snick FLNI");
	hub->command_error_nick_invalid = adc_msg_create("ISTA 221 Nick\\sinvalid");
	hub->command_error_nick_taken   = adc_msg_create("ISTA 221 Nick\\sinvalid");
	hub->command_error_cid_missing  = adc_msg_create("ISTA 243 CID\\smissing FLID");
	hub->command_error_cid_invalid  = adc_msg_create("ISTA 227 CID\\sinvalid");
	hub->command_error_cid_taken    = adc_msg_create("ISTA 224 CID\\staken");
	hub->command_error_pid_missing  = adc_msg_create("ISTA 243 PID\\smissing FLPD");
	hub->command_ban_perm           = adc_msg_create("ISTA 231 Banned");
	hub->command_ban_temp           = adc_msg_create("ISTA 232 Banned TL600"); /* FIXME: 5 minutes temp ban?! */
	hub->command_registered_only    = adc_msg_create("ISTA 226 Registered\\susers\\sonly");
	hub->command_password_wrong     = adc_msg_create("ISTA 223 Invalid\\spassword");
	hub->command_error_hub_full     = adc_msg_create("ISTA 211 Hub\\sis\\sfull");
	hub->command_error_hub_disabled = adc_msg_create("ISTA 212 Hub\\sis\\sdisabled");
	hub->command_banner             = adc_msg_create("IMSG Powered\\sby\\s" PRODUCT "/" VERSION "");
}


void hub_free_variables(struct hub_info* hub)
{
	adc_msg_free(hub->command_info);
	adc_msg_free(hub->command_banner);
	if (hub->command_motd)
		adc_msg_free(hub->command_motd);
	adc_msg_free(hub->command_support);
	adc_msg_free(hub->command_error_nick_missing);
	adc_msg_free(hub->command_error_nick_invalid);
	adc_msg_free(hub->command_error_nick_taken);
	adc_msg_free(hub->command_error_cid_missing);
	adc_msg_free(hub->command_error_cid_invalid);
	adc_msg_free(hub->command_error_cid_taken);
	adc_msg_free(hub->command_error_pid_missing);
	adc_msg_free(hub->command_ban_perm);
	adc_msg_free(hub->command_ban_temp);
	adc_msg_free(hub->command_registered_only);
	adc_msg_free(hub->command_password_wrong);
	adc_msg_free(hub->command_error_hub_full);
	adc_msg_free(hub->command_error_hub_disabled);
}



int hub_handle_info_check_nick(struct hub_info* hub, struct user* user, struct adc_message* cmd)
{
	size_t length, pos;
	int nicks;
	char* arg;
	struct user* lookup;
	
	nicks = adc_msg_has_named_argument(cmd, "NI");
	if (nicks != 1)
	{
		if (nicks == 0)
			on_login_failure(hub, user, hub->command_error_nick_missing, "nickname missing");
		else
			on_login_failure(hub, user, hub->command_error_nick_invalid, "multiple nicknames given");
		return 0;
	}
	
	arg = adc_msg_get_named_argument(cmd, "NI");
	if (arg == NULL)
		return 0;
	
	length = strlen(arg);

	/* Length of nick must at least be 2 characters long */
	if (length <= 1 || length > MAX_NICK_LEN)
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname has invalid length");
		hub_free(arg);
		return 0;
	}
	
	/* Nick must not contain newline */
	for (pos = 1; pos < length; pos++)
	{
		if (arg[pos-1] == '\\' && arg[pos] == 'n')
		{
			route_to_user(user, hub->command_error_nick_invalid);
			on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname contains newline");
			hub_free(arg);
			return 0;
		}
	}

	/* Nick must not start with a space */
	if (arg[0] == '\\' && arg[1] == 's') /* FIXME: Check for other types of spaces */
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname starts with space");
		hub_free(arg);
		return 0;
	}
	
	/*
	 * Nick should be valid utf-8, but
	 * perhaps we should check if the nick is unicode normalized?
	 */
	if (!is_valid_utf8(arg))
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname is not valid utf-8");
		hub_free(arg);
		return 0;
	}
	
	/*
	 * Since the hub sometimes send messages from itself,
	 * we do not want users named the same as the hub.
	 */
	if (!strcmp(arg, hub->config->hub_name))
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname is the same as hub name");
		hub_free(arg);
		return 0;
	}
	
	lookup = get_user_by_nick(hub, arg);
	if (lookup && lookup->sid != user->sid)
	{
		on_login_failure(hub, user, hub->command_error_nick_invalid, "nickname is already taken");
		hub_free(arg);
		return 0;
	}
	else
	{
		hub_log(log_trace, "User '%s' changed nickname to '%s'", user->nick, arg);
		
		/* Set nickname */
		memcpy(user->nick, arg, length);
		user->nick[length] = 0;
	}
	
	hub_free(arg);
	return 1;
}


int hub_handle_info_check_cid(struct hub_info* hub, struct user* user, struct adc_message* cmd)
{
	size_t pos;
	char* cid = adc_msg_get_named_argument(cmd, "ID");
	char* pid = adc_msg_get_named_argument(cmd, "PD");
	char x_pid[64];
	char raw_pid[64];
	uint64_t tiger_res[3];
	struct user* lookup = NULL;

	if (cid == NULL || pid == NULL)
	{
		hub_log(log_debug, "adc_inf_check_cid(): no cid (%p) or pid (%p)", cid, pid);
		
		if (cid == NULL)
			route_to_user(user, hub->command_error_cid_missing);
		else if (pid == NULL)
			route_to_user(user, hub->command_error_pid_missing);
		
		hub_free(cid);
		hub_free(pid);
		return 0;
	}
	
	if (strlen(cid) != strlen(pid) || strlen(cid) != MAX_CID_LEN)
	{
		route_to_user(user, hub->command_error_cid_invalid);
		hub_free(cid);
		hub_free(pid);
		return 0;
	}

	for (pos = 0; pos < MAX_CID_LEN; pos++)
	{
		if (!is_valid_base32_char(cid[pos]) || !is_valid_base32_char(pid[pos]))
		{
			route_to_user(user, hub->command_error_cid_invalid);
			hub_free(cid);
			hub_free(pid);
			return 0;
		}
	}
	
	memset(x_pid, 0, MAX_CID_LEN+1);
	base32_decode(pid, (unsigned char*) raw_pid, MAX_CID_LEN);
	tiger((uint64_t*) raw_pid, TIGERSIZE, (uint64_t*) tiger_res);
	base32_encode((unsigned char*) tiger_res, TIGERSIZE, x_pid);
	
	x_pid[MAX_CID_LEN] = 0;
	
	hub_free(pid); /* We don't need it past here */
	
	if (strncasecmp(x_pid, cid, MAX_CID_LEN) != 0)
	{
		hub_log(log_trace, "cid/pid mismatch");
		route_to_user(user, hub->command_error_cid_invalid);
		hub_free(cid);
		return 0;
	}
	
	lookup = get_user_by_cid(hub, cid);
	if (lookup && lookup->sid != user->sid)
	{
		route_to_user(user, hub->command_error_cid_taken);
		hub_free(cid);
		return 0;
	}
	else
	{
		/* Set cid */
		memcpy(user->cid, cid, MAX_CID_LEN);
		user->cid[MAX_CID_LEN] = 0;
	}

	hub_free(cid);
	return 1;
}


