/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HAVE_UHUB_HUB_H
#define HAVE_UHUB_HUB_H

struct hub_info
{
	int fd;
	struct hub_config* config;
	struct linked_list* user_list;
	sid_t free_sid;
	struct acl_handle* acl;
	struct adc_message* command_info;
	struct adc_message* command_motd;
	struct adc_message* command_support;
	struct adc_message* command_banner;
	struct adc_message* command_error_hub_full;
	struct adc_message* command_error_hub_disabled;
	struct adc_message* command_error_nick_missing;
	struct adc_message* command_error_nick_invalid;
	struct adc_message* command_error_nick_taken;
	struct adc_message* command_error_cid_missing;
	struct adc_message* command_error_cid_invalid;
	struct adc_message* command_error_cid_taken;
	struct adc_message* command_error_pid_missing;
	struct adc_message* command_ban_perm;
	struct adc_message* command_ban_temp;
	struct adc_message* command_registered_only;
	struct adc_message* command_password_wrong;
};

/**
 * This is the message pre-routing centre.
 *
 * Any message coming in to the hub comes through here first,
 * and will be routed further if valid.
 */
extern void hub_handle_message(struct hub_info* hub, struct user* u, const char* message);


/**
 * Handle protocol support/subscription messages received clients.
 */
extern void hub_handle_support(struct hub_info* hub, struct user* u, struct adc_message* cmd);

/**
 * Handle password messages received from clients.
 */
extern void hub_handle_password(struct hub_info* hub, struct user* u, struct adc_message* cmd);

/**
 * Handle info messages as received from clients.
 * This can be an initial info message, which might end up requiring password
 * authentication, etc.
 * All sorts of validation is performed here.
 * - Nickname valid?
 * - CID/PID valid?
 * - Network IP address valid?
 *
 * This can be triggered multiple times, as users can update their information,
 * in such case nickname and CID/PID changes are not allowed.
 */
extern void hub_handle_info(struct hub_info* hub, struct user* u, struct adc_message* cmd);

/**
 * Used internally by hub_handle_info
 * @return 1 if nickname is OK, or 0 if nickname is not accepted.
 */
extern int  hub_handle_info_check_nick(struct hub_info* hub, struct user* u, struct adc_message* cmd);

/**
 * Used internally by hub_handle_info
 * @return 1 if CID/PID is OK, or 0 if not valid.
 */
extern int  hub_handle_info_check_cid(struct hub_info* hub, struct user* u, struct adc_message* cmd);

/**
 * Can only be used by administrators or operators.
 */
extern void hub_handle_kick(struct hub_info* hub, struct user* u, struct adc_message* cmd);

/**
 * Send the support line for the hub to a particular user.
 * Only used during the initial handshake.
 */
extern void hub_send_support(struct hub_info* hub, struct user* u);

/**
 * Send a message assigning a SID for a user.
 * This is only sent after hub_send_support() during initial handshake.
 */
extern void hub_send_sid(struct hub_info* hub, struct user* u);

/**
 * Send a message containing hub information to a particular user.
 * This is sent during user connection, but can safely be sent at any
 * point later.
 */
extern void hub_send_hubinfo(struct hub_info* hub, struct user* u);

/**
 * Send handshake. This basically calls
 * hub_send_support() and hub_send_sid()
 */
extern void hub_send_handshake(struct hub_info* hub, struct user* u);

/**
 * Send a welcome message containing the message of the day to 
 * one particular user. This can be sent in any point in time.
 */
extern void hub_send_motd(struct hub_info* hub, struct user* u);

/**
 * Send a password challenge to a user.
 * This is only used if the user tries to access the hub using a
 * password protected nick name.
 */
extern void hub_send_password_challenge(struct hub_info* hub, struct user* u);

/**
 * Network callback to handle socket events for client sockets.
 */
extern int handle_client_network(struct net_event_listener* listener);

/**
 * Network callback to handle network events on the server socket
 * (accept clients).
 */
extern int handle_server_accept(struct net_event_listener* listener);

/**
 * This starts the hub.
 */
extern struct hub_info* hub_start_service(struct hub_config* config);

/**
 * This shuts down the hub.
 */
extern void hub_shutdown_service(struct hub_info* hub);

/**
 * This configures the hub.
 */
extern void hub_set_variables(struct hub_info* hub, struct acl_handle* acl);

/**
 * This frees the configuration of the hub.
 */
extern void hub_free_variables(struct hub_info* hub);


#endif /* HAVE_UHUB_HUB_H */

