/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"


static int server_running = 1;
static int server_restart = 0;
static int arg_verbose = 5;
static int arg_fork    = 0;
static const char* arg_uid = 0;
static const char* arg_gid = 0;
static const char* arg_config  = 0;


void hub_handle_signal(int signal)
{
	if (signal == SIGINT)
	{
		hub_log(log_info, "Interrupted. Shutting down...");
		server_running = 0;
		server_restart = 0;
	}
	else if (signal == SIGPIPE)
	{
		/* ignore it! */
		hub_log(log_trace, "hub_handle_signal(): caught SIGPIPE");
	}
	else if (signal == SIGHUP)
	{
		hub_log(log_info, "Caught hangup signal. Reloading configuration files...");
		server_running = 0;
		server_restart = 1;
	}
	else
	{
		hub_log(log_debug, "hub_handle_signal(): caught unhandled signal %d", signal);
		server_running = 0;
		server_restart = 0;
	}
}

void setup_signal_handlers()
{
	struct sigaction act;
	sigset_t sig_set;
	sigemptyset(&sig_set);
	act.sa_mask = sig_set;
	act.sa_flags = SA_ONSTACK  | SA_RESTART;
	act.sa_handler = hub_handle_signal;
	
	if (sigaction(SIGINT,  &act, 0) == -1)
		hub_log(log_error, "setup_signal_handlers(): Unable to install signal handler for SIGINT.");
	
	if (sigaction(SIGPIPE,  &act, 0) == -1)
		hub_log(log_error, "setup_signal_handlers(): Unable to install signal handler for SIGPIPE.");
		
	if (sigaction(SIGHUP,  &act, 0) == -1)
		hub_log(log_error, "setup_signal_handlers(): Unable to install signal handler for SIGHUP.");
}


int main_loop()
{
	struct hub_config configuration;
	struct acl_handle acl;
	struct hub_info* hub = 0;

	setup_signal_handlers();
	
	server_running = 1;
	server_restart = 0;

	do
	{
		if (read_config(arg_config, &configuration) == -1)
			return -1;
			
		if (acl_initialize(&configuration, &acl) == -1)
			return -1;
	
		/*
		 * Don't restart networking when re-reading configuration.
		 * This might not be possible either, since we might have
		 * dropped our privileges to do so.
		 */
		if (!server_restart)
		{
			if (net_initialize(configuration.server_max_fds) == -1)
				return -1;
		
			hub = hub_start_service(&configuration);
			if (!hub)
				return -1;
		}
		
		hub_set_variables(hub, &acl);
		
		server_running = 1;
		server_restart = 0;
		
		while (server_running)
		{
			net_wait(2500);
			user_manager_process(hub);
		}
		
		
		hub_free_variables(hub);
		
		/* Don't shutdown  networking when re-reading configuration */
		if (!server_restart)
		{
			hub_shutdown_service(hub);
			net_shutdown();
		}
		
		acl_shutdown(&acl);
		free_config(&configuration);
		
	} while(server_restart);
	
	return 0;
}


void print_version()
{
	fprintf(stdout, "" PRODUCT " " VERSION " (micro-Hub)\n");
	fprintf(stdout, "Copyright (C) 2007 Jan Vidar Krey, janvidar@extatic.org\n"
					"This is free software with ABSOLUTELY NO WARRANTY.\n\n");
	exit(0);
}


void print_usage(char* program)
{
	fprintf(stderr, "Usage: %s [options]\n\n", program);
	fprintf(stderr,
		"Options:\n"
		"   -v          Verbose mode. Add more -v's for higher verbosity.\n"
		"   -q          Quiet mode - no output\n"
		"   -f          Fork to background\n"
		"   -c <file>   Specify configuration file (default: " SERVER_CONFIG ")\n"
		"   -h          This message\n"
		"   -u <user>   Run as given user\n"
		"   -g <group>  Run with given group permissions\n"
		"   -V          Show version number.\n"
	);
	
	exit(0);
}


void parse_command_line(int argc, char** argv)
{
	int opt;
	while ((opt = getopt(argc, argv, "vqfc:hu:g:V")) != -1)
	{
		switch (opt)
		{
			case 'V':
				print_version();
				break;
			
			case 'v':
				arg_verbose++;
				break;
				
			case 'q':
				arg_verbose -= 99;
				break;
				
			case 'f':
				arg_fork = 1;
				break;
				
			case 'c':
				arg_config = optarg;
				break;
				
			case 'h':
				print_usage(argv[0]);
				break;

			case 'u':
				arg_uid = optarg;
				break;

			case 'g':
				arg_gid = optarg;
				break;

			default:
				print_usage(argv[0]);
				break;
		}
	}
	
	if (arg_config == NULL)
	{
		arg_config = SERVER_CONFIG;
	}
	
	hub_set_log_verbosity(arg_verbose);
}

int drop_privileges()
{
	struct group* perm_group = 0;
	struct passwd* perm_user = 0;
	gid_t perm_gid = 0;
	uid_t perm_uid = 0;
	int gid_ok = 0;
	int ret = 0;

	if (arg_gid)
	{
		ret = 0;
		while ((perm_group = getgrent()) != NULL)
		{
			if (strcmp(perm_group->gr_name, arg_gid) == 0)
			{
				perm_gid = perm_group->gr_gid;
				ret = 1;
				break;
			}
		}
		
		endgrent();
		
		if (!ret)
		{
			hub_log(log_fatal, "Unable to determine group id, check group name.");
			return -1;
		}

		hub_log(log_trace, "Setting group id %d (%s)", (int) perm_gid, arg_gid);
		ret = setgid(perm_gid);
		if (ret == -1)
		{
			hub_log(log_fatal, "Unable to change group id, permission denied.");
			return -1;
		}
		gid_ok = 1;
	}

	if (arg_uid)
	{
		ret = 0;
		while ((perm_user = getpwent()) != NULL)
		{
			if (strcmp(perm_user->pw_name, arg_uid) == 0)
			{
				perm_uid = perm_user->pw_uid;
				if (!gid_ok)
					perm_gid = perm_user->pw_gid;
				ret = 1;
				break;
			}
		}
		
		endpwent();
		
		if (!ret)
		{
			hub_log(log_fatal, "Unable to determine user id, check user name.");
			return -1;
		}

		if (!gid_ok) {
			hub_log(log_trace, "Setting group id %d (%s)", (int) perm_gid, arg_gid);
			ret = setgid(perm_gid);
			if (ret == -1)
			{
				hub_log(log_fatal, "Unable to change group id, permission denied.");
				return -1;
			}
		}
		
		hub_log(log_trace, "Setting user id %d (%s)", (int) perm_uid, arg_uid);
		ret = setuid(perm_uid);
		if (ret == -1)
		{
			hub_log(log_fatal, "Unable to change user id, permission denied.");
			return -1;
		}
	}
	
	return 0;
}


int main(int argc, char** argv)
{
	int ret = 0;
	
	parse_command_line(argc, argv);
	
	if (arg_fork)
	{
		ret = fork();
		if (ret == -1) 
		{
			hub_log(log_fatal, "Unable to fork to background!");
			return -1;
		}
		else if (ret == 0)
		{
			/* child process */
		}
		else
		{
			/* parent process */
			hub_log(log_debug, "Forked to background\n");
			return 0;
		}
	}
	
	if (drop_privileges() == -1)
		return -1;

	return main_loop();
}

