/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"

#ifdef DEBUG

#define UHUB_MAX_ALLOCS 5000

struct malloc_info
{
	void* ptr;
	size_t size;
	void* stack1;
	void* stack2;
};

static int    hub_alloc_count = 0;
static size_t hub_alloc_size  = 0;
static struct malloc_info hub_allocs[UHUB_MAX_ALLOCS] = { { 0, },  };

void* internal_debug_mem_malloc(size_t size, const char* where)
{
	size_t n = 0;
	char* ptr = malloc(size);
	
	if (hub_alloc_count == 0)
	{
		for (n = 0; n < UHUB_MAX_ALLOCS; n++)
		{
			hub_allocs[n].ptr    = 0;
			hub_allocs[n].size   = 0;
			hub_allocs[n].stack1 = 0;
			hub_allocs[n].stack2 = 0;
		}
	}
	
	if (ptr)
	{
		for (n = 0; n < UHUB_MAX_ALLOCS; n++)
		{
			if (!hub_allocs[n].ptr)
			{
				hub_allocs[n].ptr    = ptr;
				hub_allocs[n].size   = size;
				hub_allocs[n].stack1 = __builtin_return_address(1);
				hub_allocs[n].stack2 = __builtin_return_address(2);
				
				hub_alloc_size += size;
				hub_alloc_count++;
				
				hub_log(log_memory, "%s %p (%d bytes) (alloc: %d) (bt: %p %p)", where, ptr, (int) size, hub_alloc_count,  hub_allocs[n].stack1, hub_allocs[n].stack2);
				break;
			}
		}
	}
	else
	{
		hub_log(log_memory, "%s *** OOM for %d bytes", where, size);
		return 0;
	}
	return ptr;
}

void internal_debug_mem_free(void* ptr)
{
	if (!ptr) return;
	
	size_t n = 0;
	void* stack1 = __builtin_return_address(1);
	void* stack2 = __builtin_return_address(2);
	
	for (; n < UHUB_MAX_ALLOCS; n++)
	{
		if (hub_allocs[n].ptr == ptr)
		{
			hub_alloc_size -= hub_allocs[n].size;
			hub_alloc_count--;
			hub_allocs[n].ptr    = 0;
			hub_allocs[n].size   = 0;
			hub_allocs[n].stack1 = 0;
			hub_allocs[n].stack2 = 0;
			hub_log(log_memory, "free %p (bt: %p %p)", ptr, stack1, stack2);
			return;
		}
	}
	
	hub_log(log_memory, "free %p *** NOT ALLOCATED *** (bt: %p %p)", ptr, stack1, stack2);
}

char* debug_mem_strdup(const char* s)
{
	size_t size = strlen(s);
	char* ptr = internal_debug_mem_malloc(size+1, "strdup");
	if (ptr)
	{
		memcpy(ptr, s, size);
		ptr[size] = 0;
	}
	return ptr;
}

char* debug_mem_strndup(const char* s, size_t n)
{
	size_t size = MIN(strlen(s), n);
	char* ptr = internal_debug_mem_malloc(size+1, "strndup");
	if (ptr)
	{
		memcpy(ptr, s, size);
		ptr[size] = 0;
	}
	return ptr;
}

void* debug_mem_malloc(size_t size)
{
	void* ptr = internal_debug_mem_malloc(size, "malloc");
	return ptr;
}

void debug_mem_free(void *ptr)
{
	internal_debug_mem_free(ptr);
}



#endif


