/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"

void adc_msg_incref(struct adc_message* msg)
{
	hub_log(log_dump, "cmd: incref %p", msg);
	msg->references++;
}


void adc_msg_free(struct adc_message* msg)
{
	if (!msg)
		return;
	
	msg->references--;
	
	if (msg->references < 0)
	{
		hub_log(log_error, "cmd: free %p (DOUBLE FREE!)", msg);
		return;
	}
	
	if (msg->references == 0)
	{
		hub_log(log_dump, "cmd: free %p (free)", msg);
		hub_free(msg->cache);
		
		if (msg->feature_cast_include)
		{
			list_clear(msg->feature_cast_include, &hub_free);
			list_destroy(msg->feature_cast_include);
		}
		
		if (msg->feature_cast_exclude)
		{
			list_clear(msg->feature_cast_exclude, &hub_free);
			list_destroy(msg->feature_cast_exclude);
		}
	
		hub_free(msg);
		msg = 0;
	}
	else
	{
		hub_log(log_dump, "cmd: free %p (refcount: %d)", msg, (int) msg->references);
	}
}


struct adc_message* adc_msg_copy(struct adc_message* cmd)
{
	char* tmp = 0;
	struct adc_message* copy = (struct adc_message*) hub_malloc(sizeof(struct adc_message));
	if (!copy)
		return NULL; /* OOM */

	memcpy(copy, cmd, sizeof(struct adc_message));
	if (cmd->cache)
	{
		copy->cache = hub_malloc(cmd->length + 2);
		memcpy(copy->cache, cmd->cache, cmd->length);
		cmd->cache[cmd->length] = 0;
	}
	copy->references = 0;
	
	if (cmd->feature_cast_include)
	{
		copy->feature_cast_include = list_create();
		tmp = list_get_first(cmd->feature_cast_include);
		while (tmp)
		{
			list_append(copy->feature_cast_include, hub_strdup(tmp));
			tmp = list_get_next(cmd->feature_cast_include);
		}
	}
	
	if (cmd->feature_cast_exclude)
	{
		copy->feature_cast_exclude = list_create();
		tmp = list_get_first(cmd->feature_cast_exclude);
		while (tmp)
		{
			list_append(copy->feature_cast_exclude, hub_strdup(tmp));
			tmp = list_get_next(cmd->feature_cast_exclude);
		}
	}
	return copy;
}


struct adc_message* adc_msg_parse(struct user* u, const char* line)
{
	struct adc_message* command = (struct adc_message*) hub_malloc(sizeof(struct adc_message));
	char prefix = line[0];
	size_t length = strlen(line);
	size_t n = 0;
	char temp_sid[5];
	int ok = 1;
	struct linked_list* feature_cast_list;
	
	if (command == NULL)
		return NULL; /* OOM */

	memset(command, 0, sizeof(struct adc_message));
	command->length = length+1;
	command->capacity = command->length+1;
	command->cache = (char*) hub_malloc(command->capacity);
	
	if (command->cache == NULL)
	{
		hub_free(command);
		return NULL; /* OOM */
	}
	
	memcpy(command->cache, line, command->length);
	
	if (command->cache[length-1] == '\n')
	{
		command->length--;
	}
	else
	{
		command->cache[length - 0] = '\n';
		command->cache[length + 1] = '\0';
	}
	command->cmd = FOURCC(line[0], line[1], line[2], line[3]);
	command->priority = 0;
	command->references = 0;
	command->arg_offset = 4;
	
	adc_msg_incref(command);
	
	switch (prefix)
	{
		case 'U':
		case 'C':
			/* these should never be seen on a hub */
			ok = 0;
			break;
			
		case 'I':
		case 'H':
			ok = (length > 3);
			command->arg_offset = 4;
			break;
			
		case 'B':
			if (!u) ok = 0;
			if (!ok) break;
			
			ok = (length > 8 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			command->arg_offset = 9;
			break;
			
		case 'F':
			if (!u) ok = 0;
			if (!ok) break;
			
			ok = (length > 8 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			/* Create feature cast lists */
			command->feature_cast_include = list_create();
			command->feature_cast_exclude = list_create();
			
			if (!command->feature_cast_include || !command->feature_cast_exclude)
			{
				list_destroy(command->feature_cast_include);
				list_destroy(command->feature_cast_exclude);
				hub_free(command->cache);
				hub_free(command);
				return NULL; /* OOM */
			}
			
			n = 10;
			while (line[n] == '+' || line[n] == '-')
			{
				if (line[n++] == '+')
					feature_cast_list = command->feature_cast_include;
				else
					feature_cast_list = command->feature_cast_exclude;
				
				temp_sid[0] = line[n++];
				temp_sid[1] = line[n++];
				temp_sid[2] = line[n++];
				temp_sid[3] = line[n++];
				temp_sid[4] = '\0';
				
				list_append(feature_cast_list, hub_strdup(temp_sid));
			}
			
			command->arg_offset = n;
			
			if  (n == 10)
				ok = 0;
			
			break;
			
		case 'D':
		case 'E':
			if (!u) ok = 0;
			if (!ok) break;
			
			ok = (length > 13 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]) &&
					is_space(line[9]) &&
					is_valid_base32_char(line[10]) &&
					is_valid_base32_char(line[11]) &&
					is_valid_base32_char(line[12]) &&
					is_valid_base32_char(line[13]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			if (!ok) break;
			
			temp_sid[0] = line[10];
			temp_sid[1] = line[11];
			temp_sid[2] = line[12];
			temp_sid[3] = line[13];
			temp_sid[4] = '\0';
			
			command->target = string_to_sid(temp_sid);
			
			command->arg_offset = 14;
			break;
			
		default:
			ok = 0;
	}
	
	if (!ok) {
		adc_msg_free(command);
		return NULL;
	}
	
	/* Either 'n' is space, or end of line here! Anything else is bogus! */
	if (line[command->arg_offset] == ' ')
	{
		command->arg_offset++;
		if (line[command->arg_offset] == ' ') ok = 0;
	}
	else if (line[command->arg_offset] == '\0') ok = 1;
	else ok = 0;
	
	if (!ok)
	{
		adc_msg_free(command);
		return NULL;
	}
	
	hub_log(log_dump, "cmd: parsed %p", command);
	
	adc_msg_dump(command, "POST-PARSE");
	
	return command;
}


struct adc_message* adc_msg_create(const char* line)
{
	return adc_msg_parse(NULL, line);
}

int adc_msg_remove_named_argument(struct adc_message* cmd, const char prefix_[2])
{
	char* start;
	char* end;
	char* endInfo;
	size_t endlen;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	int found = 0;
	size_t temp_len = 0;
	
	
	adc_msg_unterminate_command(cmd);
	
	start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	while (start)
	{
		endInfo = &cmd->cache[cmd->length];
	
		if  (&start[0] < &endInfo[0])
		{
			end = strchr(&start[1], ' ');
		}
		else
		{
			end = NULL;
		}
		
		if (end)
		{
			
			temp_len = &end[0] - &start[0]; // strlen(start);
			hub_log(log_trace, "                     length=%d", (int) (temp_len));
			endlen = strlen(end);
			memmove(start, end, endlen);
			start[endlen] = '\0';
			found++;
			cmd->length -= temp_len;
			adc_msg_dump(cmd, "Removed mid arg.");
			
			
		}
		else
		{
			found++;
			cmd->length -= strlen(start);
			start[0] = '\0';
			adc_msg_dump(cmd, "Removed end arg.");
			break;
		}
		start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	}
	
	adc_msg_terminate_command(cmd);
	
	return found;
}


int adc_msg_has_named_argument(struct adc_message* cmd, const char prefix_[2])
{
	int count = 0;
	char* start;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };

	start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	while (start)
	{
		count++;
		if (strlen(start) > 1)
			start = strstr(&start[1], prefix);
		else
			start = NULL;
	}
	
	return count;
}


char* adc_msg_get_named_argument(struct adc_message* cmd, const char prefix_[2])
{
	char* start;
	char* end;
	char* argument;
	size_t length;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	
	start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	if (start == NULL)
		return NULL;
	
	start = &start[3];
	end = strchr(start, ' ');
	if (end == NULL)
		length = strlen(start);
	else
		length = &end[0] - &start[0];

	
	argument = hub_malloc(length + 1);
	if (argument == NULL)
		return NULL;

	memcpy(argument, start, length);
	argument[length] = 0;
	
	if (length > 0 && argument[length-1] == '\n')
	{
		argument[length-1] = 0;
	}
	
	return argument;
}


int adc_msg_replace_named_argument(struct adc_message* cmd, const char prefix[2], const char* string)
{
	if (adc_msg_has_named_argument(cmd, prefix))
	{
		adc_msg_remove_named_argument(cmd, prefix);
	}
	
	if (adc_msg_add_named_argument(cmd, prefix, string) == -1)
	{
		return -1;
	}
	return 0;
}


void adc_msg_terminate_command(struct adc_message* cmd)
{
	char* data = cmd->cache;
	size_t len = cmd->length;
	
	if (len > 0 && data[len-1] != '\n')
	{
		if (len >= cmd->capacity)
		{
			data = hub_malloc(len + 2); /* FIXME: OOM! */
			memcpy(data, cmd->cache, cmd->length);
			hub_free(cmd->cache);
			cmd->cache = data;
			cmd->capacity = len+2;
		}
		
		data[cmd->length++] = '\n';
		data[cmd->length] = 0;
	}
	adc_msg_dump(cmd, "TERMINATE");
}


void adc_msg_unterminate_command(struct adc_message* cmd)
{
//	adc_msg_dump(cmd, "UNTERMINATE-PRE!");
	if (cmd->length > 0 && cmd->cache[cmd->length-1] == '\n')
	{
		cmd->length--;
		cmd->cache[cmd->length] = 0;
	}
//	adc_msg_dump(cmd, "UNTERMINATE-POST");
}


int adc_msg_add_named_argument(struct adc_message* cmd, const char prefix_[2], const char* string)
{
	int ret = 0;
	char prefix[3] = { prefix_[0], prefix_[1], '\0' };
	size_t add_length = strlen(string) + 2;
	char* str = hub_malloc(add_length+1);
	
	if (!str) {
		/* FIXME: OOM error! */
		return -1;
	}
	
	str[0] = 0;
	strcat(str, prefix);
	strcat(str, string);
	
	ret = adc_msg_add_argument(cmd, str);
	hub_free(str);
	return ret;
}


int adc_msg_add_argument(struct adc_message* cmd, const char* string)
{
	char* data = 0;
	size_t len = 0;
	size_t capacity = 0;
	size_t add_length = strlen(string) + 1; /* including command separator (space) */
	hub_log(log_trace, "Add argument: '%s' (%d bytes)", string, (int) add_length);
	
	adc_msg_unterminate_command(cmd);
	
	len = cmd->length + add_length;
	if (len >= cmd->capacity)
	{
		capacity = len + 1;
		data = hub_malloc(capacity+1);
		if (data == NULL)
		{
			adc_msg_terminate_command(cmd);
			return -1;
		}
		strcpy(data, cmd->cache);
		
		hub_free(cmd->cache);
		cmd->cache = data;
		cmd->capacity = capacity;
	}
	else
	{
		data = cmd->cache;
	}
	
	strcat(data, " ");
	strcat(data, string);
	cmd->length += add_length;
	
	adc_msg_dump(cmd, "ADD ARGUMENT");
	
	adc_msg_terminate_command(cmd);
	return 0;
}


char* adc_msg_get_argument(struct adc_message* cmd, int offset)
{
	char* start;
	char* end;
	char* argument;
	int count = 0;
	
	adc_msg_unterminate_command(cmd);
	
	start = strchr(&cmd->cache[cmd->arg_offset-1], ' ');
	while (start)
	{
		end = strchr(&start[1], ' ');
		
		if (count == offset)
		{
			if (end)
			{
				argument = hub_strndup(&start[1], (&end[0] - &start[1]));
			}
			else
			{
				argument = hub_strdup(&start[1]);
				if (argument[strlen(argument)-1] == '\n')
					argument[strlen(argument)-1] = 0;
			}
			
			if (strlen(argument))
			{
				adc_msg_terminate_command(cmd);
				return argument;
			}
		}
		
		count++;
		start = end;
	}
	
	adc_msg_terminate_command(cmd);
	return 0;
}


int adc_msg_escape_length(const char* str)
{
	int add = 0;
	int n = 0;
	for (; str[n]; n++)
		if (str[n] == ' ' || str[n] == '\n' || str[n] == '\\') add++;
	return n + add;
}


int adc_msg_unescape_length(const char* str)
{
	int add = 0;
	int n = 0;
	int escape = 0;
	for (; str[n]; n++)
	{
		if (escape)
		{
			escape = 0;
		}
		else
		{
			if (str[n] == '\\')
			{
				escape = 1;
				add++;
			}
		}
	}
	return n - add;
}


char* adc_msg_unescape(const char* string)
{
	char* new_string = hub_malloc(adc_msg_unescape_length(string)+1);
	char* ptr = (char*) new_string;
	char* str = (char*) string;
	int escaped = 0;
	
	while (*str)
	{
		if (escaped) {
			if (*str == 's')
				*ptr++ = ' ';
			else if (*str == '\\')
				*ptr++ = '\\';
			else if (*str == 'n')
				*ptr++ = '\n';
			else
				*ptr++ = *str;
			
			escaped = 0;
		} else {
			if (*str == '\\')
				escaped = 1;
			else
				*ptr++ = *str;
				
		}
		str++;
	}
	*ptr = 0;
	return new_string;
}


char* adc_msg_escape(const char* string)
{
	char* str = hub_malloc(adc_msg_escape_length(string)+1);
	int n = 0;
	int i = 0;
	for (i = 0; i < strlen(string); i++)
	{
		switch (string[i]) {
			case '\\': /* fall through */
					str[n++] = '\\';
					str[n++] = '\\';
					break;
			case '\n':
					str[n++] = '\\';
					str[n++] = 'n';
					break;
			case ' ':
					str[n++] = '\\';
					str[n++] = 's';
					break;
			default:
					str[n++] = string[i];
					break;
		}
	}
	str[n] = '\0';
	return str;
}

void adc_msg_dump(struct adc_message* cmd, const char* prefix)
{
	hub_log(log_trace, "CMD_DUMP %s: '%s' (refs=%d) (%d bytes)", prefix, cmd->cache, cmd->references, strlen(cmd->cache));
	if (cmd->length != strlen(cmd->cache))
	{
		hub_log(log_error, "Command length mismatch detected: len=%d, strlen=%d, cmd='%s'", cmd->length, strlen(cmd->cache), cmd->cache);
	}
	
	if (cmd->length > cmd->capacity)
	{
		hub_log(log_error, "Command length/capacity error: len=%d, capacity=%d cmd='%s'", cmd->length, cmd->capacity, cmd->cache);
	}
	
}

