/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef __linux__

#include "uhub.h"

static int max_connections = 0;
static int num_connections = 0;
static struct epoll_event* events = 0;
static struct net_event_listener* listeners = 0;
static int epfd = -1;


static int get_position(int fd)
{
	int i;
	struct net_event_listener* listener;
	
	for (i = 0; i < max_connections; i++)
	{
		listener = &listeners[i];
		if (listener->fd == fd)
			return i;
	}
	return -1;
}


static int get_free_position()
{
	int i;
	struct net_event_listener* listener;
	
	if (num_connections >= max_connections)
		return -1;
	
	for (i = 0; i < max_connections; i++)
	{
		listener = &listeners[i];
		if (listener->fd == -1)
			return i;
	}
	
	return -1;
}


int net_initialize(int capacity)
{
	int i;
	max_connections = capacity;
	num_connections = 0;
	epfd = epoll_create(max_connections);
	if (epfd == -1)
	{
		hub_log(log_error, "net_initialize(): epoll_create failed");
		return -1;
	}
	
	events = hub_malloc(sizeof(struct epoll_event) * max_connections);
	if (!events)
	{
		hub_log(log_error, "net_initialize(): hub_malloc failed");
		return -1;
	}
	
	listeners = hub_malloc(sizeof(struct net_event_listener) * max_connections);
	if (!listeners)
	{
		hub_log(log_error, "net_initialize(): hub_malloc failed");
		hub_free(events);
		return -1;
	}
	
	memset(events, 0, sizeof(struct epoll_event) * max_connections);
	memset(listeners, 0, sizeof(struct net_event_listener) * max_connections);
	
	for (i = 0; i < max_connections; i++)
	{
		listeners[i].fd = -1;
	}
	
	net_stats_initialize();
	
	return 0;
}


int net_shutdown()
{
	hub_log(log_trace, "Shutting down network monitor");
	if (epfd != -1) {
		close(epfd);
	}
	
	hub_free(events);
	hub_free(listeners);
	
	return 0;
}


int net_wait(int timeout_ms)
{
	int fired, n, max, ret;
	struct net_event_listener* listener;
	int trig;
	
	fired = epoll_wait(epfd, events, num_connections, timeout_ms);
	if (fired == -1) {
		if (errno != EINTR)
		{
			hub_log(log_error, "net_wait(): epoll_wait failed");
		}
		return -1;
	}
	
	for (n = 0; n < fired; n++)
	{
		listener = (struct net_event_listener*) events[n].data.ptr;
		trig = events[n].events;
		
		listener->revents = 0;
		
		if (trig & EPOLLIN || trig & EPOLLPRI)
			listener->revents |= evt_read;
		
		if (trig & EPOLLOUT)
			listener->revents |= evt_write;
		
		if (trig & EPOLLHUP/* || listener->events & EPOLLRDHUP*/)
			listener->revents |= evt_close;
		
		if (trig & EPOLLERR)
			listener->revents |= evt_error;
	}
	
	max = num_connections;
	for (n = 0; n < max; n++)
	{
		listener = &listeners[n];
		if (listener && listener->fd != -1 && listener->revents != 0)
		{
			ret = listener->handler(listener);
			listener->revents = 0;
		}
	}
	
	return 0;
}


int net_add(int fd, short events, void* data, net_event_handler_t handler)
{
	struct net_event_listener* listener;
	struct epoll_event ev;
	int pos;
	
	hub_log(log_trace, "net_add(): adding socket (fd=%d)", fd);
	
	pos = get_free_position();
	if (pos == -1)
	{
		hub_log(log_error, "net_add(): unable to poll more sockets");
		return -1;
	}
	
	listener = &listeners[pos];
	listener->fd = fd;
	listener->events = events;
	listener->revents = 0;
	listener->data = data;
	listener->handler = handler;
	
	memset(&ev, 0, sizeof(struct epoll_event));
	ev.events = EPOLLIN /* | EPOLLET */;
	if (events & evt_write)
		ev.events |= EPOLLOUT;
	ev.data.ptr = listener;
	
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, fd, &ev) < 0)
	{
		hub_log(log_error, "net_add(): epoll_ctl error while adding socket (fd=%d)", fd);
		return -1;
	}
	
	num_connections++;
	return 0;
}

int net_modify(int fd, short events)
{
	struct net_event_listener* listener;
	struct epoll_event ev;
	int pos;
	hub_log(log_trace, "net_modify(): modifying socket events (fd=%d)", fd);
	
	pos = get_position(fd);
	if (pos == -1)
	{
		hub_log(log_error, "net_modify(): unable to find socket.");
		return -1;
	}
	listener = &listeners[pos];
	listener->events = events;
	
	memset(&ev, 0, sizeof(struct epoll_event));
	ev.events = EPOLLIN /* | EPOLLET */;
	if (events & evt_write)
		ev.events |= EPOLLOUT;
	ev.data.ptr = listener;
	
	if (epoll_ctl(epfd, EPOLL_CTL_MOD, fd, &ev) < 0)
	{
		hub_log(log_error, "net_add(): epoll_ctl error while modifying socket (fd=%d)", fd);
		return -1;
	}
	return 0;
}


int net_remove(int fd)
{
	struct net_event_listener* listener;
	struct epoll_event ev;
	int pos;
	
	hub_log(log_trace, "net_remove(): removing socket (fd=%d)", fd);
	
	pos = get_position(fd);
	if (pos == -1)
	{
		/* The socket is not being monitored */
		hub_log(log_error, "net_remove(): unable to remove socket (fd=%d)", fd);
		return -1;
	}
	
	listener = &listeners[pos];
	listener->fd = -1;
	listener->events = 0;
	listener->revents = 0;
	listener->data = NULL;
	listener->handler = NULL;

	if (epoll_ctl(epfd, EPOLL_CTL_DEL, fd, &ev) < 0)
	{
		hub_log(log_error, "net_remove(): epoll_ctl error while removing socket (fd=%d)", fd);
		return -1;
	}
	num_connections--;
	return 0;
}


#endif /* __linux__ */



