/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#if defined(__APPLE__) || defined(__FreeBSD__)

#include "uhub.h"

static int max_connections = 0;
static int num_connections = 0;
static struct kevent* events = 0;
static struct kevent* change = 0;
static struct net_event_listener* listeners = 0;
static int kfd = -1;


static int get_position(int fd)
{
	int i;
	struct net_event_listener* listener;
	
	for (i = 0; i < max_connections; i++)
	{
		listener = &listeners[i];
		if (listener->fd == fd)
			return i;
	}
	return -1;
}


static int get_hub_free_position()
{
	int i;
	struct net_event_listener* listener;
	
	if (num_connections >= max_connections)
		return -1;
	
	for (i = 0; i < max_connections; i++)
	{
		listener = &listeners[i];
		if (listener->fd == -1)
			return i;
	}
	
	return -1;
}


int net_initialize(int capacity)
{
	int i;
	max_connections = capacity;
	num_connections = 0;
	kfd = kqueue();
	if (kfd == -1)
	{
		hub_log(log_error, "net_initialize(): kqueue failed");
		return -1;
	}
	
	events = (void*) hub_malloc(sizeof(struct kevent) * max_connections);
	if (!events)
	{
		hub_log(log_error, "net_initialize(): hub_malloc failed");
		return -1;
	}
	
	change = (void*) hub_malloc(sizeof(struct kevent) * max_connections);
	if (!events)
	{
		hub_log(log_error, "net_initialize(): hub_malloc failed");
		hub_free(events);
		return -1;
	}
	
	
	listeners = (void*) hub_malloc(sizeof(struct net_event_listener) * max_connections);
	if (!listeners)
	{
		hub_log(log_error, "net_initialize(): hub_malloc failed");
		hub_free(change);
		hub_free(events);
		return -1;
	}
	
	memset(events, 0, sizeof(struct kevent) * max_connections);
	memset(change, 0, sizeof(struct kevent) * max_connections);
	memset(listeners, 0, sizeof(struct net_event_listener) * max_connections);
	
	for (i = 0; i < max_connections; i++)
	{
		listeners[i].fd = -1;
	}
	
	net_stats_initialize();
	
	return 0;
}


int net_shutdown()
{
	if (kfd != -1) {
		return close(kfd);
	}
	
	hub_free(events);
	hub_free(change);
	hub_free(listeners);
	return 0;
}


int net_wait(int timeout_ms)
{
	int fired, n, max, ret;
	struct net_event_listener* listener;
	int trig;
	struct timespec timeout = { (timeout_ms / 1000), (timeout_ms % 1000) * 1000 };
	
	fired = kevent(kfd, events, num_connections, change, num_connections, &timeout);
	if (fired == -1) {
		if (errno != EINTR)
		{
			hub_log(log_error, "net_wait(): kevent failed");
		}
		return -1;
	}
	
	for (n = 0; n < fired; n++)
	{
		listener = (struct net_event_listener*) change[n].udata;
		trig = change[n].flags;
		
		listener->revents = 0;
		
		if (trig & EVFILT_READ)
			listener->revents |= evt_read;
		
		if (trig & EVFILT_WRITE)
			listener->revents |= evt_write;
		
		if (trig & EV_EOF)
			listener->revents |= evt_close;
		
		if (trig & EV_ERROR)
			listener->revents |= evt_error;
	}
	
	max = num_connections;
	for (n = 0; n < max; n++)
	{
		listener = &listeners[n];
		if (listener && listener->fd != -1 && listener->revents != 0)
		{
			ret = listener->handler(listener);
			listener->revents = 0;
		}
	}
	
	return 0;
}


int net_add(int fd, short events_, void* data, net_event_handler_t handler)
{
	struct net_event_listener* listener;
	struct kevent* event;
	int pos;
	
	hub_log(log_trace, "net_add(): adding socket (fd=%d)", fd);
	
	pos = get_hub_free_position();
	if (pos == -1)
	{
		hub_log(log_error, "net_add(): unable to poll more sockets");
		return -1;
	}
	
	listener = &listeners[pos];
	listener->fd = fd;
	listener->events = events_;
	listener->revents = 0;
	listener->data = data;
	listener->handler = handler;
	
	event = &events[pos];
	event->ident = fd;
	event->filter = EVFILT_READ;
	if (events_ & evt_write)
		event->filter |= EVFILT_WRITE;
	
	event->flags = EV_ADD | EV_ENABLE;
	event->fflags = 0;
	event->data = 0;
	event->udata = listener;

	num_connections++;
	return 0;
}

int net_modify(int fd, short events_)
{
	struct net_event_listener* listener;
	struct kevent* event;
	int pos;
	
	hub_log(log_trace, "net_modify(): modifying socket (fd=%d)", fd);
	
	pos = get_position(fd);
	if (pos == -1)
	{
		/* The socket is not being monitored */
		hub_log(log_error, "net_modify(): unable to find socket (fd=%d)", fd);
		return -1;
	}
	
	listener = &listeners[pos];
	event = &events[pos];
	event->ident = fd;
	event->filter = EVFILT_READ;
	
	if (events_ & evt_write)
		event->filter |= EVFILT_WRITE;
	
	event->flags = EV_ADD | EV_ENABLE;
	
	return 0;
}


int net_remove(int fd)
{
	struct net_event_listener* listener;
	struct kevent* event;
	int pos;
	
	hub_log(log_trace, "net_remove(): removing socket (fd=%d)", fd);
	
	pos = get_position(fd);
	if (pos == -1)
	{
		/* The socket is not being monitored */
		hub_log(log_error, "net_remove(): unable to remove socket (fd=%d)", fd);
		return -1;
	}
	
	listener = &listeners[pos];
	listener->fd = -1;
	listener->events = 0;
	listener->revents = 0;
	listener->data = NULL;
	listener->handler = NULL;
	
	event = &events[pos];
	event->ident = fd;
	event->filter = 0;
	event->flags = EV_DELETE | EV_DISABLE;
	event->fflags = 0;
	event->data = 0;
	event->udata = 0;

	num_connections--;
	return 0;
}


#endif



