/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HAVE_UHUB_NETWORK_H
#define HAVE_UHUB_NETWORK_H

struct net_event_listener;

/**
 * Network event callback signature.
 * All callbacks will pass a struct net_event_listener, which
 * contains the socket descriptor and optionally a user defined data pointer.
 */
typedef int (*net_event_handler_t)(struct net_event_listener*);

enum net_events
{
	evt_none    = 0x0000,
	evt_accept  = 0x0001,
	evt_read    = 0x0002,
	evt_write   = 0x0004,
	evt_close   = 0x0008,
	evt_error   = 0x0010,
};


struct net_event_listener
{
	int fd;
	short events;
	short revents;
	void* data;
	net_event_handler_t handler;
};


struct net_statistics
{
	time_t timestamp;
	size_t tx;
	size_t rx;
	size_t accept;
	size_t closed;
	size_t errors;
};


/**
 * Initialize the socket monitor.
 * On some operating systems this will also ensure the TCP/IP stack
 * is loaded.
 *
 * @param max_connections The maximum number of sockets the monitor can handle.
 * @return -1 on error, 0 on success
 */
extern int net_initialize(int max_connections);

/**
 * Shutdown the socket monitor.
 * On some operating systems this will also ensure the TCP/IP stack
 * is loaded.
 *
 * @return -1 on error, 0 on success
 */
extern int net_shutdown();


/**
 * Add a socket to be monitored by the socket monitor.
 * 
 * @param fd socket descriptor
 * @param events the events to watch for (see enum net_events)
 * @param data a ptr to optional data associated with the socket.
 * @param handler callback function to handle the triggered socket events
 * @return -1 on error, 0 on success
 */
extern int net_add(int fd, short events, void* data, net_event_handler_t handler);

/**
 * Modify/update the trigger-events for an already monitored socket.
 *
 * @param fd socket descriptor
 * @param events the events to watch for (see enum net_events)
 * @return -1 on error, 0 on success
 */
extern int net_modify(int fd, short events);

/**
 * Stop monitoring the given socket for events.
 *
 * @param fd socket descriptor
 * @return -1 on error, 0 on success
 */
extern int net_remove(int fd);

/**
 * Wait for up to 'timeout' milli seconds for one or more network events to fire.
 * If events are fired the associated event handlers are called automatically.
 *
 * @param timeout a timeout in milli seconds
 * @return -1 on error, 0 on success
 */
extern int net_wait(int timeout);

/**
 * A wrapper for the socket() function call.
 */
extern int net_socket_create(int af, int type, int protocol);

/**
 * A wrapper for the close() function call.
 */
extern int net_close(int fd);

/**
 * A wrapper for the accept() function call.
 */
extern int net_accept(int fd);

/**
 * This will set the socket to blocking or nonblocking mode.
 * @param fd socket descriptor
 * @param toggle if non-zero nonblocking mode, otherwise blocking mode is assumed
 * @return -1 on error, 0 on success
 */
extern int net_set_nonblocking(int fd, int toggle);

/**
 * This will set or unset the SO_REUSEADDR flag.
 * @param fd socket descriptor
 * @param toggle Set SO_REUSEADDR if non-zero, otherwise unset it.
 * @return -1 on error, 0 on success
 */
extern int net_set_reuseaddress(int fd, int toggle);

/**
 * A wrapper for the recv() function call.
 */
extern ssize_t net_recv(int fd, void* buf, size_t len, int flags);

/**
 * A wrapper for the send() function call.
 */
extern ssize_t net_send(int fd, void* buf, size_t len, int flags);

/**
 * This tries to create a AF_INET6 socket.
 * If it succeeds it concludes IPv6 is supported on the host operating
 * system. If the call fails with EAFNOSUPPORT the host system
 * does not support IPv6.
 * The result is cached so further calls to this function are cheap.
 */
extern int net_is_ipv6_supported();

/**
 * This will return a string containing the peer IP-address of
 * the connected peer associated with the given socket.
 *
 * @param fd socket descriptor
 * @return IP address (IPv6 or IPv4), or "0.0.0.0" if unable to determine the address.
 */
extern const char* net_get_peer_address(int fd);


/**
 * Network statistics monitor.
 *
 * Keeps track of bandwidth usage, sockets accepted, closed,
 * errors etc.
 */
extern void net_stats_initialize();
extern void net_stats_report();
extern void net_stats_reset();
extern void net_stats_add_tx(size_t bytes);
extern void net_stats_add_rx(size_t bytes);
extern void net_stats_add_accept();
extern void net_stats_add_error();
extern void net_stats_add_close();
extern int net_stats_timeout();
extern void net_stats_get(struct net_statistics** intermediate, struct net_statistics** total);

#endif /* HAVE_UHUB_NETWORK_H */
