/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "uhub.h"


void route_message(struct hub_info* hub, struct user* u, struct adc_message* msg)
{
	adc_msg_incref(msg);
	struct user* target = NULL;
	switch (msg->cache[0]) {
		case 'B': /* Broadcast to all logged in clients */
			route_to_all(hub, msg);
			break;
			
		case 'D':
			target = get_user_by_sid(hub, msg->target);
			if (target)
			{
				route_to_user(target, msg);
			}
			break;
			
		case 'E':
			target = get_user_by_sid(hub, msg->target);
			if (target)
			{
				route_to_user(target, msg);
				route_to_user(u, msg);
			}
			break;
			
		case 'F':
			route_to_subscribers(hub, msg);
			break;	
		
		default:
			/* Ignore the message */
			break;
	}
	adc_msg_free(msg);
}


void route_to_user(struct user* user, struct adc_message* command)
{
	int ret;
	adc_msg_incref(command);
	// command->length = strlen(command->cache);
	
	adc_msg_dump(command, "Sending");
	
	if (user->send_queue_size == 0)
	{
		ret = net_send(user->sd, command->cache, command->length, UHUB_SEND_SIGNAL);
		
		if (ret == command->length)
		{
			adc_msg_free(command);
			return;
		}
		
		if (ret >= 0 || (ret == -1 && errno == EAGAIN))
		{
			list_append(user->send_queue, command);
			user->send_queue_size += command->length - ret;
			net_modify(user->sd, evt_read | evt_close | evt_write);
			
			if (ret > 0)
			{
				user->send_queue_offset = ret;
				user->tm_last_write = time(NULL);
			}
		}
		else
		{
			/* A socket error occured */
			user_disconnect(user, quit_socket_error);
			adc_msg_free(command);
		}
	}
	else
	{
		if (user->send_queue_size + command->length > user->hub->config->max_send_buffer && command->priority >= 0)
		{
			/* User is not able to swallow the data, let's cut our losses and disconnect. */
			user_disconnect(user, quit_send_queue);
			adc_msg_free(command);
		}
		else
		{
			if (user->send_queue_size + command->length > user->hub->config->max_send_buffer_soft && command->priority >= 0)
			{
				/* Don't queue this message if it is low priority! */
				adc_msg_free(command);
			}
			else
			{
				list_append(user->send_queue, command);
				user->send_queue_size += command->length;
				net_modify(user->sd, evt_read | evt_close | evt_write);
			}
		}
	}
}


void route_to_all(struct hub_info* hub, struct adc_message* command)
{
	assert(hub);
	assert(command);
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_normal)
		{
			route_to_user(user, command);
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
}


void route_to_subscribers(struct hub_info* hub, struct adc_message* command)
{
	int do_send;
	char* tmp;
	
	struct user* user = (struct user*) list_get_first(hub->user_list);
	while (user)
	{
		if (user->state == state_normal && user->feature_cast)
		{
			do_send = 1;
			
			tmp = list_get_first(command->feature_cast_include);
			while (tmp)
			{
				if (!user_have_feature_cast_support(user, tmp))
				{
					do_send = 0;
					break;
				}
				tmp = list_get_next(command->feature_cast_include);;
			}
			
			if (!do_send) {
				user = (struct user*) list_get_next(hub->user_list);
				continue;
			}
			
			tmp = list_get_first(command->feature_cast_exclude);
			while (tmp)
			{
				if (user_have_feature_cast_support(user, tmp))
				{
					do_send = 0;
					break;
				}
				tmp = list_get_next(command->feature_cast_exclude);
			}
			
			if (do_send)
			{
				route_to_user(user, command);
			}
		}
		user = (struct user*) list_get_next(hub->user_list);
	}
}

