/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HAVE_UHUB_COMMON_H
#define HAVE_UHUB_COMMON_H

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <grp.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#define PRODUCT "uHub"
#ifndef VERSION
#error "Version/build is not specified from Makefile!"
#endif

#if defined(__linux__)
#include <sys/epoll.h>
#endif

#if defined(__APPLE__) || defined(__FreeBSD__)
#include <sys/event.h>
#endif

#ifdef MSG_NOSIGNAL
#define UHUB_SEND_SIGNAL MSG_NOSIGNAL
#else
#define UHUB_SEND_SIGNAL 0
#endif


#define SERVER_PORT       1511
#define SERVER_ADDR_IPV6 "::"
#define SERVER_ADDR_IPV4 "0.0.0.0"
#define SERVER_BACKLOG 5
#define SERVER_CONFIG    "/etc/uhub/uhub.conf"
#define SERVER_ACL_FILE  "/etc/uhub/users.conf"

#define MAX_CLIENTS 512
#define MAX_CID_LEN  39
#define MAX_NICK_LEN 64
#define MAX_UA_LEN 32
#define TIGERSIZE 24

#define MAX_RECV_BUF 65535
#define MAX_ADC_CMD_LEN 4096

#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 46
#endif

typedef uint32_t sid_t;
typedef uint32_t fourcc_t;

#define FOURCC(a,b,c,d) (fourcc_t) ((a << 24) | (b << 16) | (c << 8) | d)

#define ADC_PROTO_SUPPORT "ADBASE ADBAS0"

/* Handshake and login/passwordstuff */
#define ADC_CMD_HSUP (uint32_t) (('H' << 24) | ('S' << 16) | ('U' << 8) | ('P'))
#define ADC_CMD_HPAS (uint32_t) (('H' << 24) | ('P' << 16) | ('A' << 8) | ('S'))
#define ADC_CMD_BINF (uint32_t) (('B' << 24) | ('I' << 16) | ('N' << 8) | ('F'))

/* This is a Admin extension */
#define ADC_CMD_HDSC (uint32_t) (('H' << 24) | ('D' << 16) | ('S' << 8) | ('C'))

/* Status/error messages */
#define ADC_CMD_HSTA (uint32_t) (('H' << 24) | ('S' << 16) | ('T' << 8) | ('A'))
#define ADC_CMD_DSTA (uint32_t) (('D' << 24) | ('S' << 16) | ('T' << 8) | ('A'))

/* searches */
#define ADC_CMD_BSCH (uint32_t) (('B' << 24) | ('S' << 16) | ('C' << 8) | ('H'))
#define ADC_CMD_DSCH (uint32_t) (('D' << 24) | ('S' << 16) | ('C' << 8) | ('H'))
#define ADC_CMD_ESCH (uint32_t) (('E' << 24) | ('S' << 16) | ('C' << 8) | ('H'))
#define ADC_CMD_FSCH (uint32_t) (('F' << 24) | ('S' << 16) | ('C' << 8) | ('H'))
#define ADC_CMD_DRES (uint32_t) (('D' << 24) | ('R' << 16) | ('E' << 8) | ('S'))

/* connection setup */
#define ADC_CMD_DCTM (uint32_t) (('D' << 24) | ('C' << 16) | ('T' << 8) | ('M'))
#define ADC_CMD_DRCM (uint32_t) (('D' << 24) | ('R' << 16) | ('C' << 8) | ('M'))
#define ADC_CMD_ECTM (uint32_t) (('E' << 24) | ('C' << 16) | ('T' << 8) | ('M'))
#define ADC_CMD_ERCM (uint32_t) (('E' << 24) | ('R' << 16) | ('C' << 8) | ('M'))


/* chat messages */
#define ADC_CMD_BMSG (uint32_t) (('B' << 24) | ('M' << 16) | ('S' << 8) | ('G'))
#define ADC_CMD_DMSG (uint32_t) (('D' << 24) | ('M' << 16) | ('S' << 8) | ('G'))
#define ADC_CMD_EMSG (uint32_t) (('E' << 24) | ('M' << 16) | ('S' << 8) | ('G'))
#define ADC_CMD_FMSG (uint32_t) (('F' << 24) | ('M' << 16) | ('S' << 8) | ('G'))


#define MIN(a, b) (a < b ? a : b)
#define MAX(a, b) (a > b ? a : b)

#ifdef __cplusplus
extern "C" {
#endif

#include "memory.h"
#include "misc.h"
#include "list.h"
#include "sid.h"
#include "ipcalc.h"
#include "network.h"
#include "auth.h"
#include "tiger.h"
#include "config.h"
#include "log.h"
#include "user.h"
#include "message.h"
#include "route.h"
#include "hubevent.h"
#include "hub.h"

#ifdef __cplusplus
}
#endif

#endif /* HAVE_UHUB_COMMON_H */



