/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef HAVE_UHUB_USER_H
#define HAVE_UHUB_USER_H

struct hub_info;


enum user_state
{
	state_none,                /**<< "User is connected, but has not yet sent a handshake" */
	state_protocol,            /**<< "User has sent a valid protocol handshake" */
	state_identify,            /**<< "User has sent user info data" */
	state_verify,              /**<< "User data needs to be verified by password" */
	state_normal,              /**<< "User is logged in." */
	state_wait_disconnect,     /**<< "User is disconnected, but other users need to be notified." */
	state_disconnected,        /**<< "User is disconnected" */
};


enum user_credentials
{
	cred_none,                 /**<<< "User has no credentials (ERROR STATE!)" */
	cred_link,                 /**<<< "User is a link (not used currently)" */
	cred_user,                 /**<<< "User is a regular user (guest)" */
 	cred_user_registered,      /**<<< "User is identified as a registered user" */
	cred_operator,             /**<<< "User is identified as a hub operator" */
	cred_admin,                /**<<< "User is identified as a hub administrator" */
};


enum user_features
{
	feature_base   = 0x00000001, /**<<< "Basic configuration (required by all clients)" */
	feature_auto   = 0x00000002, /**<<< "Automatic nat detection traversal" */
	feature_bbs    = 0x00000004, /**<<< "Bulletin board system (not supported by this software)" */
	feature_ucmd   = 0x00000008, /**<<< "User commands (not supported by this software)" */
	feature_zlif   = 0x00000010, /**<<< "gzip stream compression (not supported by this software)" */
};


enum user_quit_reason
{
	quit_disconnected,           /**<<< "User disconnected" */
	quit_kicked,                 /**<<< "User was kicked" */
	quit_banned,                 /**<<< "User was banned" */
	quit_timeout,                /**<<< "User timed out (no data for a while)" */
	quit_send_queue,             /**<<< "User's send queue was overflowed" */
	quit_memory_error,           /**<<< "Not enough memory available" */
	quit_socket_error,           /**<<< "A socket error occured" */
	quit_protocol_error,         /**<<< "Fatalprotocol error" */
	quit_logon_error,            /**<<< "Unable to login (wrong password, CID/PID, etc)" */
};


struct user
{
	int                  sd;                      /**<<< "socket descriptor" */
	int                  state;                   /**<<< "see enum user_state" */
	int                  credentials;             /**<<< "see enum user_credentials" */
	sid_t                sid;                     /**<<< "ADC session ID" */
	char                 cid[MAX_CID_LEN+1];      /**<<< "ADC global client ID" */
	char                 nick[MAX_NICK_LEN+1];    /**<<< "ADC nick name" */
	char                 user_agent[MAX_UA_LEN+1];/**<<< "User agent string" */
	time_t               tm_connected;            /**<<< "time when user connected" */
	time_t               tm_last_read;            /**<<< "time the user last received something from the hub" */
	time_t               tm_last_write;           /**<<< "time the user last sent something to the hub" */
	int                  features;                /**<<< "see enum user_features" */
	struct linked_list*  feature_cast;            /**<<< "Features supported by feature cast" */
	struct adc_message*  info;                    /**<<< "ADC 'INF' command" */
	size_t               send_queue_offset;       /**<<< "Send queue byte offset" */
	struct linked_list*  send_queue;              /**<<< "Send queue" */
	int                  send_queue_size;         /**<<< "Size of send queue (in bytes, not messages)" */
	char*                recv_buf;                /**<<< "Recv buffer" */
	size_t               recv_buf_offset;         /**<<< "Recv buffer offset" */
	struct hub_info*     hub;                     /**<<< "The hub instance this user belong to" */
	int                  quit_reason;             /**<<< "Quit reason (see user_quit_reason)" */
};

/**
 * Initializes the user manager.
 * @return 0 on success, or -1 if error (out of memory).
 */
extern int user_manager_init(struct hub_info* hub);

/**
 * Shuts down the user manager.
 * All users will be disconnected and deleted as part of this.
 */
extern void user_manager_shutdown(struct hub_info* hub);

/**
 * Process users. This will timeout stale connections and
 * perform scheduled cleanups of users.
 * This should be called periodically.
 *
 * FIXME: Make sure we don't do double iterations of the user database
 * in case we call user_destroy().
 */
extern void user_manager_process(struct hub_info* hub);

/**
 * Create a user with the given socket descriptor.
 * @param sd socket descriptor associated with the user
 * @return User object or NULL if not enough memory is available.
 */
extern struct user* user_create(struct hub_info* hub, int sd);

/**
 * Delete a user. If the user is logged in a quit message is issued.
 */
extern void user_destroy(struct user* user);

/**
 * This associates a INF message to the user.
 * If the user already has a INF message associated, then this is
 * released before setting the new one.
 */
extern void user_set_info(struct user* user, struct adc_message* info);

/**
 * User supports the protocol extension as given in fourcc.
 * This is usually set while the user is connecting, but can
 * also be used to subscribe to a new class of messages from the
 * hub.
 *
 * @see enum user_features
 */
extern void user_support_add(struct user* user, int fourcc);

/**
 * User no longer supports the protocol extension as given in fourcc.
 * This can be used to unsubscribe to certain messages generated by
 * the hub.
 * @see enum user_features
 */
extern void user_support_remove(struct user* user, int fourcc);

/**
 * Check if a user supports 'feature' for feature casting (basis for 'Fxxx' messages)
 * The feature cast is specified as the 'SU' argument to the user's
 * INF-message.
 * 
 * @param feature a feature to lookup (example: 'TCP4' or 'UDP4')
 * @return 1 if 'feature' supported, or 0 otherwise
 */
extern int user_have_feature_cast_support(struct user* user, char feature[4]);

/**
 * Lookup a user based on the session ID (sid).
 * @return a user if found, or NULL if not found
 */
extern struct user* get_user_by_sid(struct hub_info* hub, sid_t sid);

/**
 * Lookup a user based on the client ID (cid).
 * @return a user if found, or NULL if not found
 */
extern struct user* get_user_by_cid(struct hub_info* hub, const char* cid);

/**
 * Lookup a user based on the nick name.
 * @return a user if found, or NULL if not found
 */
extern struct user* get_user_by_nick(struct hub_info* hub, const char* nick);


/**
 * Send the user list of connected clients to 'user'.
 */
extern void send_user_list(struct hub_info* hub, struct user* user);

/**
 * Send a quit message for 'user' (leaving).
 */
extern void send_quit_message(struct hub_info* hub, struct user* user);

/**
 * Disconnect the user.
 * 
 * @param user User to disconnect
 * @param reason See enum user_quit_reason
 */
extern void user_disconnect(struct user* user, int reason);

/**
 * Service the send queue for 'user'.
 * FIXME: This should not be here!
 */
extern void send_queued_data(struct user* user);



#endif /* HAVE_UHUB_USER_H */


